import unittest
from unittest.mock import MagicMock, patch

import pandas as pd

from petsard.evaluator.anonymeter import Anonymeter


class TestAnonymeter(unittest.TestCase):
    """Test for Anonymeter Evaluator."""

    def setUp(self):
        """Set up test fixtures."""
        # Test configuration
        self.config = {
            "eval_method": "anonymeter-singlingout",
            "n_attacks": 100,
            "n_cols": 2,
            "max_attempts": 1000,
            "mode": "multivariate",
        }

        # Create mock data
        self.data = {
            "ori": pd.DataFrame(
                {
                    "col1": [1, 2, 3, 4, 5],
                    "col2": ["a", "b", "c", "d", "e"],
                    "col3": [1.1, 2.2, 3.3, 4.4, 5.5],
                }
            ),
            "syn": pd.DataFrame(
                {
                    "col1": [1, 2, 3, 4, 5],
                    "col2": ["a", "b", "c", "d", "e"],
                    "col3": [1.1, 2.2, 3.3, 4.4, 5.5],
                }
            ),
            "control": pd.DataFrame(
                {
                    "col1": [6, 7, 8, 9, 10],
                    "col2": ["f", "g", "h", "i", "j"],
                    "col3": [6.6, 7.7, 8.8, 9.9, 10.0],
                }
            ),
        }

    @patch("anonymeter.evaluators.SinglingOutEvaluator")
    def test_init(self, mock_evaluator):
        """Test initialization."""
        evaluator = Anonymeter(config=self.config)
        self.assertEqual(evaluator.config, self.config)
        self.assertIsNone(evaluator._impl)

    @patch("anonymeter.evaluators.SinglingOutEvaluator")
    def test_eval_singlingout(self, mock_singlingout):
        """Test eval method with SinglingOut."""
        # Setup mock returns
        mock_instance = MagicMock()
        mock_instance.risk.return_value.value = 0.5
        mock_instance.risk.return_value.ci = [0.4, 0.6]
        mock_instance.results.return_value.attack_rate.value = 0.3
        mock_instance.results.return_value.attack_rate.error = 0.1
        mock_instance.results.return_value.baseline_rate.value = 0.2
        mock_instance.results.return_value.baseline_rate.error = 0.05
        mock_instance.results.return_value.control_rate.value = 0.1
        mock_instance.results.return_value.control_rate.error = 0.02
        mock_singlingout.return_value = mock_instance

        # Execute evaluator
        evaluator = Anonymeter(config=self.config)
        result = evaluator.eval(self.data)

        # Assert results structure
        self.assertIn("global", result)
        self.assertIn("details", result)

        # Assert global results content
        global_data = result["global"]
        self.assertIsInstance(global_data, pd.DataFrame)
        result_dict = global_data.iloc[0].to_dict()
        self.assertIn("risk", result_dict)
        self.assertIn("risk_CI_btm", result_dict)
        self.assertIn("risk_CI_top", result_dict)
        self.assertIn("attack_rate", result_dict)
        self.assertEqual(result_dict["risk"], 0.5)  # Check specific value

    @patch("anonymeter.evaluators.LinkabilityEvaluator")
    def test_eval_linkability(self, mock_linkability):
        """Test eval method with Linkability."""
        # Setup config for linkability
        linkability_config = {
            "eval_method": "anonymeter-linkability",
            "n_attacks": 100,
            "n_neighbors": 5,
            "aux_cols": [["col1"], ["col3"]],
            "n_jobs": -1,
        }

        # Setup mock returns
        mock_instance = MagicMock()
        mock_instance.risk.return_value.value = 0.4
        mock_instance.risk.return_value.ci = [0.3, 0.5]
        mock_instance.results.return_value.attack_rate.value = 0.25
        mock_instance.results.return_value.attack_rate.error = 0.08
        mock_instance.results.return_value.baseline_rate.value = 0.15
        mock_instance.results.return_value.baseline_rate.error = 0.04
        mock_instance.results.return_value.control_rate.value = 0.1
        mock_instance.results.return_value.control_rate.error = 0.02
        mock_linkability.return_value = mock_instance

        # Execute evaluator
        evaluator = Anonymeter(config=linkability_config)
        result = evaluator.eval(self.data)

        # Assert structure and content
        self.assertIn("global", result)
        self.assertIsInstance(result["global"], pd.DataFrame)
        self.assertEqual(result["global"].iloc[0]["risk"], 0.4)

    def test_invalid_method(self):
        """Test with invalid evaluation method."""
        with self.assertRaises(Exception):
            Anonymeter(config={"eval_method": "anonymeter-invalid"})


if __name__ == "__main__":
    unittest.main()



"""Pure functions for field-level operations"""

import re
from typing import Any, Callable, Dict, Optional, Tuple

import numpy as np
import pandas as pd
from pandas.api.types import (
    is_float_dtype,
    is_integer_dtype,
    is_numeric_dtype,
    is_object_dtype,
)

from petsard.metadater.field.field_types import FieldConfig, FieldMetadata, FieldStats
from petsard.metadater.types.data_types import DataType, LogicalType


def build_field_metadata(
    field_data: pd.Series,
    field_name: str,
    config: Optional[FieldConfig] = None,
    compute_stats: bool = True,
    infer_logical_type: bool = True,
    optimize_dtype: bool = True,
    sample_size: Optional[int] = 1000,
) -> FieldMetadata:
    """
    Pure function to build FieldMetadata from a pandas Series

    Args:
        field_data: The pandas Series to analyze
        field_name: Name of the field
        config: Optional field configuration
        compute_stats: Whether to compute statistics
        infer_logical_type: Whether to infer logical type
        optimize_dtype: Whether to optimize dtype
        sample_size: Sample size for analysis

    Returns:
        FieldMetadata object
    """
    if config is None:
        config = FieldConfig()

    # Determine data type
    pandas_dtype = _safe_dtype_string(field_data.dtype)
    data_type = _map_pandas_to_metadater_type(pandas_dtype)

    # 檢查前導零情況，如果發現前導零則強制設為字串類型
    if config.auto_detect_leading_zeros:
        if (
            data_type in [DataType.INT8, DataType.INT16, DataType.INT32, DataType.INT64]
            or pandas_dtype == "object"
        ):
            if _has_leading_zeros(field_data):
                data_type = DataType.STRING

    # Override with type hint if provided
    if config.type_hint:
        data_type = _apply_type_hint(config.type_hint, data_type)

    # Determine nullable
    nullable = (
        config.nullable if config.nullable is not None else field_data.isna().any()
    )

    # Build base metadata
    field_metadata = FieldMetadata(
        name=field_name,
        data_type=data_type,
        nullable=nullable,
        source_dtype=pandas_dtype,
        description=config.description,
        cast_error=config.cast_error,
        properties=config.properties.copy(),
    )

    # Infer logical type
    if infer_logical_type:
        if config.logical_type:
            try:
                logical_type = LogicalType(config.logical_type)
                field_metadata = field_metadata.with_logical_type(logical_type)
            except ValueError:
                # Invalid logical type, infer from data
                logical_type = infer_field_logical_type(field_data, field_metadata)
                if logical_type:
                    field_metadata = field_metadata.with_logical_type(logical_type)
        else:
            logical_type = infer_field_logical_type(field_data, field_metadata)
            if logical_type:
                field_metadata = field_metadata.with_logical_type(logical_type)

    # Calculate statistics
    if compute_stats:
        sample_data = field_data
        if sample_size and len(field_data) > sample_size:
            sample_data = field_data.sample(n=sample_size, random_state=42)
        stats = calculate_field_stats(sample_data, field_metadata)
        field_metadata = field_metadata.with_stats(stats)

    # Determine optimal target dtype
    if optimize_dtype:
        target_dtype = optimize_field_dtype(field_data, field_metadata, config)
        field_metadata = field_metadata.with_target_dtype(target_dtype)

    return field_metadata


def calculate_field_stats(
    field_data: pd.Series, field_metadata: FieldMetadata
) -> FieldStats:
    """
    Pure function to calculate field statistics

    Args:
        field_data: The pandas Series to analyze
        field_metadata: Field metadata for context

    Returns:
        FieldStats object
    """
    stats = FieldStats(
        row_count=len(field_data),
        na_count=int(field_data.isna().sum()),
    )

    # Calculate na_percentage
    na_percentage = (
        (stats.na_count / stats.row_count) * 100 if stats.row_count > 0 else 0.0
    )
    stats = FieldStats(
        row_count=stats.row_count,
        na_count=stats.na_count,
        na_percentage=round(na_percentage, 4),
        distinct_count=int(field_data.nunique()),
    )

    # For numeric types, calculate additional stats
    if field_metadata.data_type in [
        DataType.INT8,
        DataType.INT16,
        DataType.INT32,
        DataType.INT64,
        DataType.FLOAT32,
        DataType.FLOAT64,
    ]:
        if not field_data.dropna().empty:
            quantiles = {
                0.25: field_data.quantile(0.25),
                0.5: field_data.quantile(0.5),
                0.75: field_data.quantile(0.75),
            }

            stats = FieldStats(
                row_count=stats.row_count,
                na_count=stats.na_count,
                na_percentage=stats.na_percentage,
                distinct_count=stats.distinct_count,
                min_value=field_data.min(),
                max_value=field_data.max(),
                mean_value=float(field_data.mean()),
                std_value=float(field_data.std()),
                quantiles=quantiles,
            )

    # Most frequent values
    value_counts = field_data.value_counts().head(10)
    if not value_counts.empty:
        most_frequent = list(zip(value_counts.index, value_counts.values))
        stats = FieldStats(
            row_count=stats.row_count,
            na_count=stats.na_count,
            na_percentage=stats.na_percentage,
            distinct_count=stats.distinct_count,
            min_value=stats.min_value,
            max_value=stats.max_value,
            mean_value=stats.mean_value,
            std_value=stats.std_value,
            quantiles=stats.quantiles,
            most_frequent=most_frequent,
        )

    return stats


def infer_field_logical_type(
    field_data: pd.Series, field_metadata: FieldMetadata
) -> Optional[LogicalType]:
    """
    Pure function to infer logical type from data patterns

    Args:
        field_data: The pandas Series to analyze
        field_metadata: Field metadata for context

    Returns:
        Inferred LogicalType or None
    """
    # Only process string-like data
    if field_metadata.data_type not in [DataType.STRING, DataType.BINARY]:
        return None

    sample = field_data.dropna()
    if len(sample) == 0:
        return None

    # Limit sample size
    if len(sample) > 1000:
        sample = sample.sample(n=1000, random_state=42)

    # Ensure string operations are possible
    if not _is_string_compatible(sample):
        return None

    try:
        if not pd.api.types.is_string_dtype(sample):
            sample = sample.astype(str)
    except Exception:
        return None

    # Check patterns
    patterns = _get_logical_type_patterns()
    for logical_type, (pattern, threshold) in patterns.items():
        try:
            matches = sample.str.match(pattern, na=False)
            match_ratio = matches.sum() / len(matches)

            if match_ratio >= threshold:
                # Special validation for certain types
                if logical_type in [
                    LogicalType.LATITUDE,
                    LogicalType.LONGITUDE,
                    LogicalType.PERCENTAGE,
                    LogicalType.CURRENCY,
                ]:
                    validator = _get_special_validator(logical_type)
                    if not _validate_with_function(sample, validator, threshold):
                        continue
                return logical_type
        except Exception:
            continue

    # Check categorical
    unique_ratio = field_data.nunique() / len(field_data)
    if unique_ratio < 0.05:  # 5% threshold
        return LogicalType.CATEGORICAL

    return None


def optimize_field_dtype(
    field_data: pd.Series,
    field_metadata: FieldMetadata,
    config: Optional[FieldConfig] = None,
) -> str:
    """
    Pure function to determine optimal dtype for storage

    Args:
        field_data: The pandas Series to analyze
        field_metadata: Field metadata for context
        config: Optional field configuration

    Returns:
        Optimal dtype string
    """
    if config is None:
        config = FieldConfig()

    if (
        field_metadata.data_type == DataType.STRING
        and field_metadata.logical_type == LogicalType.CATEGORICAL
    ):
        return "category"

    # 使用完整的類型分析邏輯
    return _comprehensive_type_analysis(field_data, config)


# Helper functions (pure)


def _safe_dtype_string(dtype: Any) -> str:
    """Convert various dtype representations to string"""
    if isinstance(dtype, np.dtype):
        return dtype.name
    elif isinstance(dtype, pd.CategoricalDtype):
        return f"category[{dtype.categories.dtype.name}]"
    elif isinstance(dtype, pd.api.extensions.ExtensionDtype):
        return str(dtype)
    elif isinstance(dtype, str):
        return dtype
    elif isinstance(dtype, type):
        return dtype.__name__.lower()
    else:
        return str(dtype)


def _map_pandas_to_metadater_type(pandas_dtype: str) -> DataType:
    """Convert pandas dtype to Metadater DataType"""
    dtype_str = pandas_dtype.lower()

    mapping = {
        "bool": DataType.BOOLEAN,
        "int8": DataType.INT8,
        "int16": DataType.INT16,
        "int32": DataType.INT32,
        "int64": DataType.INT64,
        "uint8": DataType.INT16,
        "uint16": DataType.INT32,
        "uint32": DataType.INT64,
        "uint64": DataType.INT64,
        "float16": DataType.FLOAT32,
        "float32": DataType.FLOAT32,
        "float64": DataType.FLOAT64,
        "object": DataType.STRING,
        "string": DataType.STRING,
        "category": DataType.STRING,
    }

    if dtype_str.startswith("datetime64"):
        if "utc" in dtype_str.lower() or "tz" in dtype_str.lower():
            return DataType.TIMESTAMP_TZ
        return DataType.TIMESTAMP

    if dtype_str.startswith("category"):
        return DataType.STRING

    return mapping.get(dtype_str, DataType.STRING)


def _apply_type_hint(type_hint: str, current_type: DataType) -> DataType:
    """Apply type hint to determine data type"""
    type_hint = type_hint.lower()
    type_mapping = {
        "category": DataType.STRING,
        "datetime": DataType.TIMESTAMP,
        "date": DataType.DATE,
        "time": DataType.TIME,
        "int": DataType.INT64,
        "integer": DataType.INT64,
        "float": DataType.FLOAT64,
        "string": DataType.STRING,
        "boolean": DataType.BOOLEAN,
    }
    return type_mapping.get(type_hint, current_type)


def _is_string_compatible(series: pd.Series) -> bool:
    """Check if series can use string operations"""
    if pd.api.types.is_string_dtype(series):
        return True
    if pd.api.types.is_object_dtype(series):
        try:
            non_null = series.dropna()
            if len(non_null) == 0:
                return False
            return all(isinstance(x, str) for x in non_null.head(100))
        except Exception:
            return False
    return False


def _get_logical_type_patterns() -> Dict[LogicalType, Tuple[str, float]]:
    """Get logical type detection patterns with confidence thresholds"""
    return {
        LogicalType.EMAIL: (r"^[\w\.-]+@[\w\.-]+\.\w+$", 0.95),
        LogicalType.URL: (r"^https?://[^\s]+$", 0.95),
        LogicalType.IP_ADDRESS: (
            r"^(?:(?:25[0-5]|2[0-4][0-9]|[01]?[0-9][0-9]?)\.){3}(?:25[0-5]|2[0-4][0-9]|[01]?[0-9][0-9]?)$",
            0.95,
        ),
        LogicalType.UUID: (
            r"^[0-9a-fA-F]{8}-[0-9a-fA-F]{4}-[0-9a-fA-F]{4}-[0-9a-fA-F]{4}-[0-9a-fA-F]{12}$",
            0.98,
        ),
    }


def _get_special_validator(logical_type: LogicalType) -> Callable[[str], bool]:
    """Get special validator function for logical type"""
    validators = {
        LogicalType.LATITUDE: _validate_latitude,
        LogicalType.LONGITUDE: _validate_longitude,
        LogicalType.PERCENTAGE: _validate_percentage,
        LogicalType.CURRENCY: _validate_currency,
    }
    return validators[logical_type]


def _validate_with_function(
    sample: pd.Series, validator: Callable[[str], bool], threshold: float
) -> bool:
    """Validate sample using a custom validator function"""
    try:
        valid_count = sum(validator(x) for x in sample)
        return (valid_count / len(sample)) >= threshold
    except Exception:
        return False


def _validate_latitude(value: str) -> bool:
    """Validate latitude"""
    try:
        lat = float(value)
        return -90 <= lat <= 90
    except (ValueError, TypeError):
        return False


def _validate_longitude(value: str) -> bool:
    """Validate longitude"""
    try:
        lon = float(value)
        return -180 <= lon <= 180
    except (ValueError, TypeError):
        return False


def _validate_percentage(value: str) -> bool:
    """Validate percentage"""
    if value.endswith("%"):
        try:
            float(value[:-1])
            return True
        except ValueError:
            return False
    try:
        num = float(value)
        return 0 <= num <= 1 or 0 <= num <= 100
    except ValueError:
        return False


def _validate_currency(value: str) -> bool:
    """Validate currency"""
    currency_patterns = [
        r"^\$[\d,]+\.?\d*$",
        r"^[\d,]+\.?\d*\$$",
        r"^€[\d,]+\.?\d*$",
        r"^£[\d,]+\.?\d*$",
        r"^¥[\d,]+\.?\d*$",
        r"^[A-Z]{3}\s?[\d,]+\.?\d*$",
    ]
    return any(re.match(pattern, value) for pattern in currency_patterns)


def _is_float_with_integer_values(field_data: pd.Series) -> bool:
    """
    檢測 float 序列是否實際上都是整數值

    Args:
        field_data: 要檢查的 Series

    Returns:
        True 如果所有非空值都是整數，False 否則
    """
    if not is_float_dtype(field_data):
        return False

    # 移除空值
    clean_data = field_data.dropna()
    if len(clean_data) == 0:
        return False

    # 檢查所有值是否都是整數
    try:
        return all(val.is_integer() for val in clean_data)
    except (AttributeError, TypeError):
        return False


def _comprehensive_type_analysis(
    field_data: pd.Series, config: Optional[FieldConfig] = None
) -> str:
    """
    完整的類型分析邏輯，按照以下順序：
    1. 檢查前導零 -> 字串
    2. 檢查小數點 -> float
    3. 檢查整數（含空值處理）-> int/Int64
    4. 其他 -> category/string

    Args:
        field_data: 要分析的 Series
        config: 欄位配置

    Returns:
        最佳的 dtype 字串
    """
    if config is None:
        config = FieldConfig()

    # 1. 首先檢查前導零（無論原始類型為何）
    if config.auto_detect_leading_zeros and _has_leading_zeros(field_data):
        return "string"

    # 2. 統一嘗試數值轉換（無論原始類型為何）
    try:
        # 嘗試轉換為數值（保留 NaN）
        numeric_data = pd.to_numeric(field_data, errors="coerce")

        # 計算有效數值的比例
        non_na_count = field_data.notna().sum()
        valid_numeric_count = numeric_data.notna().sum()

        if non_na_count > 0 and valid_numeric_count > 0:
            valid_numeric_ratio = valid_numeric_count / non_na_count

            if valid_numeric_ratio >= 0.8:  # 80% 以上可以轉為數值
                valid_numeric_values = numeric_data.dropna()

                if len(valid_numeric_values) > 0:
                    # 檢查是否都是整數值
                    # 需要處理 int 和 float 兩種情況
                    try:
                        if is_integer_dtype(valid_numeric_values):
                            # 如果已經是整數類型，直接認為是整數
                            all_integers = True
                        else:
                            # 如果是浮點類型，檢查是否都是整數值
                            all_integers = all(
                                val.is_integer() for val in valid_numeric_values
                            )
                    except (AttributeError, TypeError):
                        # 如果無法判斷，假設不是整數
                        all_integers = False

                    if all_integers:
                        # 整數資料，檢查是否有空值
                        if field_data.isna().any() and config.force_nullable_integers:
                            return _optimize_nullable_integer_dtype(
                                valid_numeric_values
                            )
                        else:
                            return _optimize_regular_integer_dtype(valid_numeric_values)
                    else:
                        # 浮點數資料
                        return _optimize_float_dtype(numeric_data)

    except Exception:
        pass

    # 如果數值轉換失敗，但原本就是數值類型，使用原來的邏輯
    if is_numeric_dtype(field_data):
        return _optimize_numeric_dtype(field_data, config)

    # 3. 如果不是數值資料，檢查其他類型
    if is_object_dtype(field_data):
        # 對於 object 類型，先檢查是否為 datetime，否則返回 category
        clean_data = field_data.dropna()
        if len(clean_data) == 0:
            return "category"

        # 檢查是否為 datetime
        if _might_be_datetime(clean_data):
            try:
                datetime_data = pd.to_datetime(clean_data, errors="coerce")
                if not datetime_data.isna().any():
                    return "datetime64[s]"
            except Exception:
                pass

        return "category"
    elif is_numeric_dtype(field_data):
        # 如果原本就是數值類型，使用原來的邏輯
        return _optimize_numeric_dtype(field_data, config)

    # 其他情況保持原樣
    return str(field_data.dtype)


def _optimize_float_dtype(numeric_data: pd.Series) -> str:
    """優化浮點數類型"""
    if numeric_data.isna().all():
        return "float32"

    col_min, col_max = np.nanmin(numeric_data), np.nanmax(numeric_data)

    # 檢查是否 float32 足夠
    if np.finfo(np.float32).min <= col_min and col_max <= np.finfo(np.float32).max:
        return "float32"
    else:
        return "float64"


def _optimize_nullable_integer_dtype(integer_data: pd.Series) -> str:
    """優化 nullable integer 類型"""
    if len(integer_data) == 0:
        return "Int64"

    col_min, col_max = integer_data.min(), integer_data.max()

    ranges = [
        ("Int8", np.iinfo(np.int8).min, np.iinfo(np.int8).max),
        ("Int16", np.iinfo(np.int16).min, np.iinfo(np.int16).max),
        ("Int32", np.iinfo(np.int32).min, np.iinfo(np.int32).max),
        ("Int64", np.iinfo(np.int64).min, np.iinfo(np.int64).max),
    ]

    for dtype, min_val, max_val in ranges:
        if min_val <= col_min and col_max <= max_val:
            return dtype
    return "Int64"


def _optimize_regular_integer_dtype(integer_data: pd.Series) -> str:
    """優化一般 integer 類型"""
    if len(integer_data) == 0:
        return "int64"

    col_min, col_max = integer_data.min(), integer_data.max()

    ranges = [
        ("int8", np.iinfo(np.int8).min, np.iinfo(np.int8).max),
        ("int16", np.iinfo(np.int16).min, np.iinfo(np.int16).max),
        ("int32", np.iinfo(np.int32).min, np.iinfo(np.int32).max),
        ("int64", np.iinfo(np.int64).min, np.iinfo(np.int64).max),
    ]

    for dtype, min_val, max_val in ranges:
        if min_val <= col_min and col_max <= max_val:
            return dtype
    return "int64"


def _optimize_numeric_dtype(
    field_data: pd.Series, config: Optional[FieldConfig] = None
) -> str:
    """Optimize numeric dtype"""
    if config is None:
        config = FieldConfig()

    if is_integer_dtype(field_data):
        # 如果有空值且啟用 force_nullable_integers，使用 nullable integer 避免轉為 float
        if field_data.isna().any() and config.force_nullable_integers:
            if field_data.isna().all():
                return "Int64"

            col_min, col_max = np.nanmin(field_data), np.nanmax(field_data)

            # 使用 nullable integer types
            ranges = {
                "Int8": (np.iinfo(np.int8).min, np.iinfo(np.int8).max),
                "Int16": (np.iinfo(np.int16).min, np.iinfo(np.int16).max),
                "Int32": (np.iinfo(np.int32).min, np.iinfo(np.int32).max),
                "Int64": (np.iinfo(np.int64).min, np.iinfo(np.int64).max),
            }

            for dtype, (min_val, max_val) in ranges.items():
                if min_val <= col_min and col_max <= max_val:
                    return dtype
            return "Int64"
        else:
            # 沒有空值時使用傳統 integer types
            col_min, col_max = np.nanmin(field_data), np.nanmax(field_data)

            ranges = {
                "int8": (np.iinfo(np.int8).min, np.iinfo(np.int8).max),
                "int16": (np.iinfo(np.int16).min, np.iinfo(np.int16).max),
                "int32": (np.iinfo(np.int32).min, np.iinfo(np.int32).max),
                "int64": (np.iinfo(np.int64).min, np.iinfo(np.int64).max),
            }

            for dtype, (min_val, max_val) in ranges.items():
                if min_val <= col_min and col_max <= max_val:
                    return dtype
            return "int64"

    elif is_float_dtype(field_data):
        # 檢查是否實際上是整數值
        if _is_float_with_integer_values(field_data):
            # 轉換為 nullable integer
            if field_data.isna().all():
                return "Int64"

            col_min, col_max = np.nanmin(field_data), np.nanmax(field_data)

            ranges = {
                "Int8": (np.iinfo(np.int8).min, np.iinfo(np.int8).max),
                "Int16": (np.iinfo(np.int16).min, np.iinfo(np.int16).max),
                "Int32": (np.iinfo(np.int32).min, np.iinfo(np.int32).max),
                "Int64": (np.iinfo(np.int64).min, np.iinfo(np.int64).max),
            }

            for dtype, (min_val, max_val) in ranges.items():
                if min_val <= col_min and col_max <= max_val:
                    return dtype
            return "Int64"
        else:
            # 真正的浮點數
            if field_data.isna().all():
                return "float32"

            col_min, col_max = np.nanmin(field_data), np.nanmax(field_data)

            ranges = {
                "float32": (np.finfo(np.float32).min, np.finfo(np.float32).max),
                "float64": (np.finfo(np.float64).min, np.finfo(np.float64).max),
            }

            for dtype, (min_val, max_val) in ranges.items():
                if min_val <= col_min and col_max <= max_val:
                    return dtype
            return "float64"
    else:
        return str(field_data.dtype)


def _might_be_datetime(field_data: pd.Series) -> bool:
    """
    Pre-check if field data might contain datetime values to avoid unnecessary parsing warnings.

    Args:
        field_data: Clean pandas Series (no NaN values)

    Returns:
        True if data might be datetime, False otherwise
    """
    if len(field_data) == 0:
        return False

    # Sample a few values to check
    sample_size = min(10, len(field_data))
    sample = field_data.head(sample_size)

    # Convert to string for pattern checking
    try:
        sample_str = sample.astype(str)
    except Exception:
        return False

    datetime_indicators = 0

    # First, check for obvious non-datetime patterns that should be excluded
    non_datetime_patterns = [
        "hs-grad",
        "some-college",
        "bachelors",
        "masters",
        "doctorate",
        "prof-school",
        "assoc-acdm",
        "assoc-voc",
        "11th",
        "10th",
        "9th",
        "7th-8th",
        "5th-6th",
        "machine-op-inspct",
        "farming-fishing",
        "protective-serv",
        "other-service",
        "never-married",
        "married-civ-spouse",
        "divorced",
        "separated",
        "widowed",
        "own-child",
        "husband",
        "wife",
        "not-in-family",
        "unmarried",
        "united-states",
        "asian-pac-islander",
        "amer-indian-eskimo",
    ]

    # Check if any sample values match common non-datetime patterns
    for value in sample_str:
        value_lower = value.strip().lower()
        if any(pattern in value_lower for pattern in non_datetime_patterns):
            return False  # Definitely not datetime data

    for value in sample_str:
        value = value.strip()

        # Skip very short strings (likely not datetime)
        if len(value) < 4:
            continue

        # Check for common datetime patterns
        # Date separators: -, /, .
        if any(sep in value for sep in ["-", "/", "."]):
            # Check if contains digits (years, months, days)
            if any(char.isdigit() for char in value):
                # Additional check: must have at least 4 consecutive digits (likely a year)
                # or multiple digit groups separated by separators
                import re

                digit_groups = re.findall(r"\d+", value)
                if len(digit_groups) >= 2 or any(
                    len(group) >= 4 for group in digit_groups
                ):
                    datetime_indicators += 1
                    continue

        # Check for time patterns (contains :)
        if ":" in value and any(char.isdigit() for char in value):
            datetime_indicators += 1
            continue

        # Check for month names
        month_names = [
            "jan",
            "feb",
            "mar",
            "apr",
            "may",
            "jun",
            "jul",
            "aug",
            "sep",
            "oct",
            "nov",
            "dec",
            "january",
            "february",
            "march",
            "april",
            "may",
            "june",
            "july",
            "august",
            "september",
            "october",
            "november",
            "december",
        ]
        if any(month in value.lower() for month in month_names):
            datetime_indicators += 1
            continue

    # Require at least 70% of sampled values to have strong datetime indicators
    # and ensure we actually checked some values
    if sample_size == 0:
        return False
    return datetime_indicators / sample_size >= 0.7


def _has_leading_zeros(field_data: pd.Series) -> bool:
    """
    檢測字串是否包含前導零的數字

    Args:
        field_data: 要檢查的 Series

    Returns:
        True 如果發現前導零模式，False 否則
    """
    if field_data.isna().all():
        return False

    field_data_clean = field_data.dropna()
    if len(field_data_clean) == 0:
        return False

    # 轉換為字串進行檢查
    try:
        field_data_str = field_data_clean.astype(str)
    except Exception:
        return False

    # 檢查前導零模式：開頭是 "0" 且後面跟數字
    leading_zero_count = 0
    total_checked = 0

    for value in field_data_str.head(100):  # 檢查前100個值以提高效能
        value = value.strip()
        if len(value) >= 2:  # 至少要有兩個字符
            # 檢查是否符合前導零模式：0開頭且全部都是數字
            if value.startswith("0") and value.isdigit():
                leading_zero_count += 1
            total_checked += 1

    # 如果超過30%的值有前導零，則認為這是前導零欄位
    if total_checked > 0:
        return (leading_zero_count / total_checked) >= 0.3

    return False


def _optimize_object_dtype(
    field_data: pd.Series, config: Optional[FieldConfig] = None
) -> str:
    """Optimize object dtype"""
    if config is None:
        config = FieldConfig()

    if field_data.isna().all():
        return "category"

    field_data_clean = field_data.dropna()

    # 首先檢查是否有前導零，如果有則保持為字串
    if config.auto_detect_leading_zeros and _has_leading_zeros(field_data):
        return "string"

    # Pre-check if data might be datetime before attempting conversion
    # This avoids unnecessary datetime parsing warnings for obviously non-datetime data
    if _might_be_datetime(field_data_clean):
        # Try datetime conversion
        try:
            datetime_field_data = pd.to_datetime(field_data_clean, errors="coerce")
            if not datetime_field_data.isna().any():
                return "datetime64[s]"
        except Exception:
            pass

    return "category"



from dataclasses import dataclass

from petsard.evaluator.data_describer import DataDescriber
from petsard.evaluator.evaluator import Evaluator, EvaluatorConfig, EvaluatorMap
from petsard.evaluator.evaluator_base import BaseEvaluator
from petsard.exceptions import UnsupportedMethodError


@dataclass
class DescriberConfig(EvaluatorConfig):
    """
    Configuration for the describer.
        Inherits from EvaluatorConfig.

    Attributes:
        _logger (logging.Logger): The logger object.
        DEFAULT_EVALUATING_METHOD (str): The default method for evaluating the data.
        DEFAULT_DESCRIBING_METHOD (str): The default method for describing the data.
        method (str): The method to be used for evaluating the data.
        method_code (int): The code of the evaluator method.
        eval_method (str): The name of the evaluator method.
            The difference between 'method' and 'eval_method' is that 'method' is the user input,
            while 'eval_method' is the actual method used for evaluating the data
        custom_params (dict): Any additional parameters to be stored in custom_params.

    """

    DEFAULT_DESCRIBING_METHOD: str = "describe"

    def __post_init__(self):
        super().__post_init__()

    def _init_eval_method(self) -> None:
        """
        Initialize the eval_method attribute based on the method_code.
            Overridden EvaluatorConfig's _init_eval_method().
        """
        error_msg: str = None

        try:
            method_code: int = EvaluatorMap.map(self.method.lower())

            # Filter method from EvaluatorMap.
            #   Remember 'default' in Describer means 'describe',
            #   but 'default' in Evaluator means 'sdmetrics-single_table-qualityreport'
            if method_code not in (EvaluatorMap.DEFAULT, EvaluatorMap.DESCRIBE):
                error_msg = f"Unsupported describer method: {self.method}"
                self._logger.error(error_msg)
                raise UnsupportedMethodError(error_msg)

            self.method_code: int = method_code
            self._logger.debug(
                f"Mapped evaluating method '{self.method}' to code {self.method_code}"
            )
        except KeyError:
            error_msg = f"Unsupported describer method: {self.method}"
            self._logger.error(error_msg)
            raise UnsupportedMethodError(error_msg)

        # Set the default
        self.eval_method: str = (
            self.DEFAULT_DESCRIBING_METHOD
            if self.method_code == EvaluatorMap.DEFAULT
            else self.method
        )
        self._logger.info(
            f"DescriberConfig initialized with method: {self.method}, eval_method: {self.eval_method}"
        )


class Describer(Evaluator):
    """
    The Describer class is responsible generate statistical summaries
    and insights from datasets.
    """

    DESCIRBER_MAP: dict[int, BaseEvaluator] = {
        EvaluatorMap.DEFAULT: DataDescriber,
        EvaluatorMap.DESCRIBE: DataDescriber,
    }

    def __init__(self, method: str, **kwargs):
        """
        Args:
            method (str): The method to be used for evaluating the data.
            **kwargs: Any additional parameters to be stored in custom_params.

        Attr:
            _logger (logging.Logger): The logger object.
            config (DescriberConfig): The configuration parameters for the describer.
            _impl (BaseEvaluator): The evaluator object.
        """
        super().__init__(method, **kwargs)

    def _configure_implementation(self, method: str, **kwargs) -> None:
        """
        Configure the describer's implementation based on the specified method.
            Overridden Evaluator's _configure_implementation().

        Args:
            method (str): The method to be used for describe the data.
            **kwargs: Any additional parameters to be stored in custom_params.
        """
        # Initialize the EvaluatorConfig object
        self.config: DescriberConfig = DescriberConfig(method=method)
        self._logger.debug("DescriberConfig successfully initialized")

        # Add custom parameters to the config
        if kwargs:
            self._logger.debug(
                f"Additional keyword arguments provided: {list(kwargs.keys())}"
            )
            self.config.update({"custom_params": kwargs})
            self._logger.debug("Config successfully updated with custom parameters")
        else:
            self._logger.debug("No additional parameters provided")

    def _create_evaluator_class(self) -> BaseEvaluator:
        """
        Create a Evaluator object with the configuration parameters.
            Overridden Evaluator's _create_evaluator_class().

        Returns:
            BaseEvaluator: The evaluator object.
        """
        return self.DESCIRBER_MAP[self.config.method_code]
        return self.DESCIRBER_MAP[self.config.method_code]
        return self.DESCIRBER_MAP[self.config.method_code]



import logging
import re
import warnings
from dataclasses import dataclass, field
from enum import Enum, auto
from typing import Any

import pandas as pd
from anonymeter.evaluators import (
    InferenceEvaluator,
    LinkabilityEvaluator,
    SinglingOutEvaluator,
)

from petsard.config_base import BaseConfig
from petsard.evaluator.evaluator_base import BaseEvaluator
from petsard.exceptions import ConfigError, UnsupportedMethodError
from petsard.metadater.types.data_types import safe_round


class AnonymeterMap(Enum):
    """
    map of Anonymeter
    """

    SINGLINGOUT: int = auto()
    LINKABILITY: int = auto()
    INFERENCE: int = auto()

    @classmethod
    def map(cls, method: str) -> int:
        """
        Get suffixes mapping int value

        Args:
            method (str): evaluating method

        Return:
            (int): The method code.
        """
        return cls.__dict__[re.sub(r"^anonymeter-", "", method).upper()]


@dataclass
class AnonymeterConfig(BaseConfig):
    """
    Configuration for the Anonymeter Evaluator.

    Attributes:
        _logger (logging.Logger): The logger object.
        eval_method (str): The method of how you evaluating data.
        eval_method_code (int, optional): The code of the evaluator method.
        AVAILABLE_MODES (list[str]): The available modes for SinglingOut.
        n_attacks (int, optional):
            The number of attack attempts using the specified attack method.
        max_n_attacks (bool, optional):
            Determines whether to enforce the maximum number of attacks.
            Support only for Linkability and Inference.
        max_attempts (int, optional):
            The maximum number of singling-out attempts to find a valid query.
        n_cols (int, optional):
            The number of columns used for generating singling-out queries.
        mode (str, optional): The mode of singling-out evaluation.
        n_neighbors (int, optional):
            Sets the number of nearest neighbors to consider for each entry in the search.
            Indicating a successful linkability attack only if
            the closest synthetic record matches for both split original records.
        aux_cols (tuple[list[str], list[str]], optional):
            Features of the records that are given to the attacker as auxiliary information.
            The Anonymeter documentation states it supports 'tuple of int',
            but this is not reflected in their type annotations,
            so we will omit it here and only mention this for reference.
        secret (str | list[str], optional):
            The secret attribute(s) of the target records, unknown to the attacker.
            This is what the attacker will try to guess.
        n_jobs (int, optional):
            Specifies the number of jobs Anonymeter will use. Not supported for SinglingOut.
            -1 means all threads except one. -2 means every thread.
        ori (pd.DataFrame): The original data.
        syn (pd.DataFrame): The synthetic data.
        control (pd.DataFrame): The control data.
    """

    eval_method: str
    eval_method_code: int = None

    AVAILABLE_MODES: list[str] = field(
        default_factory=lambda: ["multivariate", "univariate"]
    )

    n_attacks: int = 2000
    max_n_attacks: bool = False
    max_attempts: int = 500000
    n_cols: int = 3
    mode: str = "multivariate"
    n_neighbors: int = 1
    aux_cols: tuple[list[str], list[str]] = None
    secret: str | list[str] = None
    n_jobs: int = -2
    ori: pd.DataFrame = None
    syn: pd.DataFrame = None
    control: pd.DataFrame = None

    def __post_init__(self):
        super().__post_init__()
        error_msg: str = None

        try:
            self.eval_method_code = AnonymeterMap.map(self.eval_method)
        except KeyError:
            error_msg = f"Unsupported evaluator method: {self.eval_method}"
            self._logger.error(error_msg)
            raise UnsupportedMethodError(error_msg)

        if self.n_attacks <= 0:
            error_msg = "The number of attacks must be greater than 0."
            self._logger.error(error_msg)
            raise ConfigError(error_msg)

        if self.n_jobs == 0 or self.n_jobs < -2:
            error_msg = "The number of jobs must be -2, -1, or greater than 0."
            self._logger.error(error_msg)
            raise ConfigError(error_msg)

        if self.eval_method_code == AnonymeterMap.SINGLINGOUT:
            if self.mode not in self.AVAILABLE_MODES:
                error_msg = f"Unsupported mode for SinglingOut: {self.mode}"
                self._logger.error(error_msg)
                raise ConfigError(error_msg)
            elif self.mode == "univariate":
                self._logger.info(
                    f"n_cols setting {self.n_cols} will be ignored "
                    "due to SinglingOut mode set to 'univariate'"
                )
                self.n_cols = None
            else:  # multivariate
                if self.n_cols <= 0:
                    error_msg = "The number of singling-out attack columns must be greater than 0."
                    self._logger.error(error_msg)
                    raise ConfigError(error_msg)

            if self.max_attempts <= 0:
                error_msg = "The maximum number of attempts must be greater than 0."
                self._logger.error(error_msg)
                raise ConfigError(error_msg)

            self.n_neighbors = None
            self.aux_cols = None
            self.secret = None
            self.n_jobs = None
        elif self.eval_method_code in [
            AnonymeterMap.LINKABILITY,
            AnonymeterMap.INFERENCE,
        ]:
            self.max_attempts = None
            self.n_cols = None
            self.mode = None

            if self.max_n_attacks:
                self._logger.info(
                    "max_n_attacks is set, the maximum number, 'n_attacks' will be adjusted"
                )

            if self.eval_method_code == AnonymeterMap.LINKABILITY:
                if self.aux_cols is None:
                    error_msg = (
                        "The auxiliary columns must be provided for Linkability."
                    )
                    self._logger.error(error_msg)
                    raise ConfigError(error_msg)
                elif (
                    len(self.aux_cols) != 2
                    or not (
                        isinstance(self.aux_cols[0], list)
                        and isinstance(self.aux_cols[1], list)
                    )
                    or not all(
                        isinstance(item, str)
                        for item in self.aux_cols[0] + self.aux_cols[1]
                    )
                ):
                    error_msg = (
                        "The auxiliary columns must be a tuple of two lists of strings."
                    )
                    self._logger.error(error_msg)
                    raise ConfigError(error_msg)
                elif set(self.aux_cols[0]).intersection(set(self.aux_cols[1])):
                    overlapping_columns: set[str] = set(self.aux_cols[0]).intersection(
                        set(self.aux_cols[1])
                    )
                    error_msg = (
                        f"The two lists in auxiliary columns must not have overlapping elements. "
                        f"Found overlapping elements: {overlapping_columns}"
                    )
                    self._logger.error(error_msg)
                    raise ConfigError(error_msg)
            else:  # Inference
                if self.aux_cols is not None:
                    if not isinstance(self.aux_cols, list) or not all(
                        isinstance(item, str) for item in self.aux_cols
                    ):
                        error_msg = "The auxiliary columns must be a list of strings."
                        self._logger.error(error_msg)
                        raise ConfigError(error_msg)

                    if self.secret is None:
                        error_msg = (
                            "The secret attribute(s) must be provided for Inference."
                        )
                        self._logger.error(error_msg)
                        raise ConfigError(error_msg)
                    else:
                        if self.secret in self.aux_cols:
                            error_msg = "The secret attribute(s) must not be included in the auxiliary columns."
                            self._logger.error(error_msg)
                            raise ConfigError(error_msg)

    def update_data(self, data: dict) -> None:
        error_msg: str = None

        self.ori = data["ori"]
        self.syn = data["syn"]
        self.control = data["control"]

        if self.eval_method_code == AnonymeterMap.SINGLINGOUT:
            if self.n_cols > self.ori.shape[1]:
                error_msg = (
                    "The singling must be less than the number of columns in the data."
                )
                self._logger.error(error_msg)
                raise ConfigError(error_msg)

        elif self.eval_method_code in [
            AnonymeterMap.LINKABILITY,
            AnonymeterMap.INFERENCE,
        ]:
            if self.max_n_attacks:
                ori_n_attacks: int = self.n_attacks
                self.max_n_attacks = self.control[0]
                self._logger.info(
                    f"Adjusted 'n_attacks' from {ori_n_attacks} to {self.n_attacks} "
                    "due to 'max_n_attacks'"
                )

            if self.eval_method_code == AnonymeterMap.INFERENCE:
                if self.aux_cols is None:
                    self.aux_cols = [
                        col for col in self.ori.columns if col != self.secret
                    ]

            all_aux_cols: list[str] = (
                self.aux_cols[0] + self.aux_cols[1]
                if self.eval_method_code == AnonymeterMap.LINKABILITY
                else self.aux_cols
            )
            invalid_columns: list[str] = [
                col for col in all_aux_cols if col not in self.ori.columns
            ]
            if invalid_columns:
                error_msg = f"The following auxiliary columns do not exist in the original data: {invalid_columns}"
                self._logger.error(error_msg)
                raise ConfigError(error_msg)

            if self.eval_method_code == AnonymeterMap.INFERENCE:
                if self.secret not in self.ori.columns:
                    error_msg = f"The secret attribute(s) '{self.secret}' must exist in the original data."
                    self._logger.error(error_msg)
                    raise ConfigError(error_msg)


class Anonymeter(BaseEvaluator):
    """
    Factory for Anonymeter Evaluator.
    """

    REQUIRED_INPUT_KEYS: list[str] = ["ori", "syn", "control"]
    ANONYMETER_CLASS_MAP: dict[int, Any] = {
        AnonymeterMap.SINGLINGOUT: SinglingOutEvaluator,
        AnonymeterMap.LINKABILITY: LinkabilityEvaluator,
        AnonymeterMap.INFERENCE: InferenceEvaluator,
    }
    REQUIRED_ANONYMETER_KEYS: list[str] = [
        "ori",
        "syn",
        "control",
        "n_attacks",
    ]
    REQUIRED_ANONYMETER_KEYS_MAP: dict[int, list[str]] = {
        AnonymeterMap.SINGLINGOUT: (
            REQUIRED_ANONYMETER_KEYS + ["n_cols", "max_attempts"]
        ),
        AnonymeterMap.LINKABILITY: (
            REQUIRED_ANONYMETER_KEYS + ["n_neighbors", "aux_cols"]
        ),
        AnonymeterMap.INFERENCE: (REQUIRED_ANONYMETER_KEYS + ["secret", "aux_cols"]),
    }
    MAX_ATTEMPTS_OF_RUNTIME_ERROR: int = 10
    REQUIRED_ANONYMETER_KEYS_EVALUATE_MAP: dict[int, list[str]] = {
        AnonymeterMap.SINGLINGOUT: ["mode"],
        AnonymeterMap.LINKABILITY: ["n_jobs"],
        AnonymeterMap.INFERENCE: ["n_jobs"],
    }
    AVAILABLE_SCORES_GRANULARITY: list[str] = [
        "global",
        "details",
    ]

    def __init__(self, config: dict):
        """
        Args:
            config (dict): A dictionary containing the configuration settings.
                - eval_method (str): The method of how you evaluating data.

        Attr:
            REQUIRED_INPUT_KEYS (list[str]): The required keys for the input data.
            ANONYMETER_CLASS_MAP (dict[int, Any]): The mapping of the Anonymeter classes.
            REQUIRED_ANONYMETER_KEYS (list[str]): The required keys for the Anonymeter.
            REQUIRED_ANONYMETER_KEYS_MAP (dict[int, list[str]]):
            MAX_ATTEMPTS_OF_RUNTIME_ERROR (int): The maximum number of attempts for runtime error.
            REQUIRED_ANONYMETER_KEYS_EVALUATE_MAP (dict[int, list[str]]):
            AVAILABLE_SCORES_GRANULARITY (list[str]): The available scores granularity.
            _logger (logging.Logger): The logger object.
            config (dict): A dictionary containing the configuration settings.
            anonymeter_config (AnonymeterConfig):
                The configuration parameters for the anonymeter.
            _impl (Any): The evaluator object.
        """
        super().__init__(config=config)
        self._logger: logging.Logger = logging.getLogger(
            f"PETsARD.{self.__class__.__name__}"
        )

        self._logger.debug(f"Verify AnonymeterConfig with parameters: {self.config}")
        self.anonymeter_config = AnonymeterConfig(**self.config)
        self._logger.debug(
            "AnonymeterConfig successfully initialized as {self.anonymeter_config}"
        )

        self._impl: Any = None

    def _extract_scores(self) -> dict[str, Any]:
        """
        _extract_scores of Anonymeter.
            Uses .risk()/.results() method in Anonymeter
            to extract result from self._impl into the designated dictionary.

        Return
            (dict[str, Any]). Result as specific format describe in eval().
        """

        result: dict[str, Any] = {}

        # Handle the risk
        try:
            risk = self._impl.risk()
            result["risk"] = safe_round(risk.value)
            result["risk_CI_btm"] = safe_round(risk.ci[0])
            result["risk_CI_top"] = safe_round(risk.ci[1])
        except Exception:
            result["risk"] = pd.NA
            result["risk_CI_btm"] = pd.NA
            result["risk_CI_top"] = pd.NA

        # Handle the attack_rate, baseline_rate, control_rate
        try:
            results = self._impl.results()
            for rate_type in ["attack_rate", "baseline_rate", "control_rate"]:
                rate_result = getattr(results, rate_type, None)
                if rate_result:
                    result[f"{rate_type}"] = safe_round(rate_result.value)
                    result[f"{rate_type}_err"] = safe_round(rate_result.error)
                else:
                    result[f"{rate_type}"] = pd.NA
                    result[f"{rate_type}_err"] = pd.NA
        except Exception:
            for rate_type in ["attack_rate", "baseline_rate", "control_rate"]:
                result[f"{rate_type}"] = pd.NA
                result[f"{rate_type}_err"] = pd.NA
            error_msg: str = "There's no result for attack_rate, recorded as NA."
            self._logger.warning(error_msg)

        return result

    def _get_global(self, anonymeter_scores: dict[str, pd.DataFrame]) -> pd.DataFrame:
        """
        Retrieves the global result from the Anonymeter.

        Args:
            anonymeter_scores (dict[str, pd.DataFrame]): The scores of the Anonymeter.

        Returns:
            (pd.DataFrame): A DataFrame with the global evaluation result.
                One row only for representing the whole data result.
        """

        return pd.DataFrame.from_dict(
            data={"result": anonymeter_scores}, orient="columns"
        ).T

    def _get_details(
        self, anonymeter_scores: dict[str, pd.DataFrame]
    ) -> dict[str, Any]:
        """
        Retrieves the specify details of the evaluation.

        Args:
            anonymeter_scores (dict[str, pd.DataFrame]): The scores of the Anonymeter.

        Returns:
            (dict[str, Any]): The details of the evaluation.
        """
        if self.anonymeter_config.eval_method_code == AnonymeterMap.INFERENCE:
            return None  # Inference queries didn't been stored

        details: dict[str, Any] = {}

        # SinglingOut attacks of Univariate. control queries didn't been stored
        if self.anonymeter_config.eval_method_code == AnonymeterMap.SINGLINGOUT:
            details["attack_queries"] = self._impl.queries(baseline=False)
            details["baseline_queries"] = self._impl.queries(baseline=True)
        # Linkability: Dict[int, Set(int)], aux_cols[0] indexes links to aux_cols[1]
        elif self.anonymeter_config.eval_method_code == AnonymeterMap.LINKABILITY:
            n_neighbors = self.anonymeter_config.n_neighbors
            details["attack_links"] = self._impl._attack_links.find_links(
                n_neighbors=n_neighbors
            )
            details["baseline_links"] = self._impl._baseline_links.find_links(
                n_neighbors=n_neighbors
            )
            details["control_links"] = self._impl._control_links.find_links(
                n_neighbors=n_neighbors
            )

        return details

    def _eval(self, data: dict[str, pd.DataFrame]) -> dict[str, pd.DataFrame]:
        """
        Evaluating the evaluator.
            _impl should be initialized in this method.

        Args:
            data (dict[str, pd.DataFrame]): The data to be evaluated.

        Returns:
            (dict) Result as following key-value pairs:
            - risk (float)
                Privacy Risk value of specified attacks. Ranging from 0 to 1.
                    A value of 0 indicates no risk, and 1 indicates full risk.
                Includes risk_ci_btm and risk_ci_top
                    for the bottom and top of the confidence interval.
            - attack_Rate (float)
                Main attack rate of specified attacks, which the attacker
                    uses the synthetic datase to deduce private information of records
                    in the original/training dataset. Ranging from 0 to 1.
                A value of 0 indicates none of success attack,
                    and 1 indicates totally success attack.
                Includes attack_Rate_err for its error rate.
            - baseline_Rate (float)
                Naive, or Baseline attack rate of specified attacks, which is
                    carried out based on random guessing, to provide a baseline against
                    which the strength of the “main” attack can be compared.
                    Ranging from 0 to 1.
                A value of 0 indicates none of success attack,
                    and 1 indicates totally success attack.
                Includes baseline_Rate_err for its error rate.
            - control_Rate (float)
                Control attack rate of specified attacks,  hich is conducted on
                    a set of control dataset, to distinguish the concrete privacy risks
                    of the original data records (i.e., specific information)
                    from general risks intrinsic to the whole population (i.e., generic information).
                    Ranging from 0 to 1.
                A value of 0 indicates none of success attack,
                    and 1 indicates totally success attack.
                Includes control_Rate_err for its error rate.
        """

        self.anonymeter_config.update_data(data)

        self._logger.debug(
            f"Initializing evaluator with method: {self.config['eval_method']}"
        )
        evaluator_class: Any = self.ANONYMETER_CLASS_MAP[
            self.anonymeter_config.eval_method_code
        ]
        self._logger.debug(
            f"Mapped method code: {self.anonymeter_config.eval_method_code}"
        )

        self._impl = evaluator_class(
            **self.anonymeter_config.get_params(
                param_configs=[
                    {attr: {"action": "INCLUDE"}}
                    for attr in self.REQUIRED_ANONYMETER_KEYS_MAP[
                        self.anonymeter_config.eval_method_code
                    ]
                ]
            )
        )

        # catch warnings during synthesizer initialization:
        # UserWarning: Attack is as good or worse as baseline model.
        # FutureWarning: is_categorical_dtype is deprecated and will be removed in a future version.
        with warnings.catch_warnings(record=True) as w:
            warnings.simplefilter("always")

            attempts_for_impl_evaluate: int = 0
            while attempts_for_impl_evaluate < self.MAX_ATTEMPTS_OF_RUNTIME_ERROR:
                attempts_for_impl_evaluate += 1
                try:
                    self._impl.evaluate(
                        **self.anonymeter_config.get_params(
                            param_configs=[
                                {attr: {"action": "INCLUDE"}}
                                for attr in self.REQUIRED_ANONYMETER_KEYS_EVALUATE_MAP[
                                    self.anonymeter_config.eval_method_code
                                ]
                            ]
                        )
                    )
                    break  # successful
                except RuntimeError:
                    error_msg: str = (
                        f"Evaluating trial {attempts_for_impl_evaluate} / {self.MAX_ATTEMPTS_OF_RUNTIME_ERROR}"
                        " failed due to RunTimeError."
                    )
                    self._logger.warning(error_msg)

                    if attempts_for_impl_evaluate == self.MAX_ATTEMPTS_OF_RUNTIME_ERROR:
                        error_msg = f"Exceeded the maximum number of attempts ({self.MAX_ATTEMPTS_OF_RUNTIME_ERROR}). "
                        self._logger.warning(error_msg)

            for warning in w:
                self._logger.debug(f"Warning during _eval: {warning.message}")

        # self._extract_scores() already handle the exception by assign NA
        anonymeter_scores: dict[str, pd.DataFrame] = self._extract_scores()
        self._logger.debug(f"Extracted scores: {list(anonymeter_scores.keys())}")

        scores: dict[str, pd.DataFrame] = {}
        for granularity in self.AVAILABLE_SCORES_GRANULARITY:
            self._logger.debug(f"Extracting {granularity} level as PETsARD format")

            if granularity == "global":
                scores[granularity] = self._get_global(anonymeter_scores)
            elif granularity == "details":
                scores[granularity] = self._get_details(anonymeter_scores)
        self._logger.info("Successfully extracting scores")

        return scores



import logging
from dataclasses import dataclass, field
from enum import Enum, auto
from typing import Any, Optional

import pandas as pd

from petsard.config_base import BaseConfig
from petsard.evaluator.data_describer_base import (
    DataDescriberColNA,
    DataDescriberColumnCount,
    DataDescriberCorr,
    DataDescriberCov,
    DataDescriberGlobalNA,
    DataDescriberKurtosis,
    DataDescriberMax,
    DataDescriberMean,
    DataDescriberMedian,
    DataDescriberMin,
    DataDescriberNUnique,
    DataDescriberPercentile,
    DataDescriberQ1,
    DataDescriberQ3,
    DataDescriberRowCount,
    DataDescriberSkew,
    DataDescriberStd,
    DataDescriberVar,
)
from petsard.evaluator.evaluator_base import BaseEvaluator
from petsard.exceptions import ConfigError, UnsupportedMethodError
from petsard.metadater import safe_round


class DataDescriberMap(Enum):
    """
    Mapping of the describing method to the corresponding code.
    """

    ROW_COUNT: int = auto()
    COL_COUNT: int = auto()
    GLOBAL_NA_COUNT: int = auto()
    MENA: int = auto()
    MEDIAN: int = auto()
    STD: int = auto()
    VAR: int = auto()
    MIN: int = auto()
    MAX: int = auto()
    KURTOSIS: int = auto()
    SKEW: int = auto()
    Q1: int = auto()
    Q3: int = auto()
    IQR: int = auto()
    RANGE: int = auto()
    PERCENTILE: int = auto()
    COL_NA_COUNT: int = auto()
    NUNIUQE: int = auto()
    COV: int = auto()
    CORR: int = auto()

    @classmethod
    def map(cls, method: str) -> int:
        """
        Get suffixes mapping int value

        Args:
            method (str): evaluating method

        Return:
            (int): The method code.
        """
        return cls.__dict__[method.upper()]


@dataclass
class DataDescriberConfig(BaseConfig):
    """
    Configuration for the Stats Evaluator.

    Attributes:
        eval_method (str): The evaluation method.
        eval_method_code (int): The evaluation method code.
        describe_method (list[str]): The list of describing methods.
    """

    eval_method: str
    eval_method_code: Optional[int] = None

    AVAILABLE_DESCRIBE_METHODS: list[str] = field(
        default_factory=lambda: [
            "row_count",
            "col_count",
            "global_na_count",
            "mean",
            "median",
            "std",
            "var",
            "min",
            "max",
            "kurtosis",
            "skew",
            "q1",
            "q3",
            "percentile",
            "col_na_count",
            "nunique",
            "corr",
            "cov",
        ]
    )
    REQUIRED_INPUT_KEYS: list[str] = field(default_factory=lambda: ["data"])

    describe_method: str | list[str] = field(
        default_factory=lambda: [
            "row_count",
            "col_count",
            "global_na_count",
            "mean",
            "median",
            "std",
            "min",
            "max",
            "kurtosis",
            "skew",
            "q1",
            "q3",
            "col_na_count",
            "nunique",
            "corr",
        ]
    )
    percentile: int | float = None

    def __post_init__(self):
        super().__post_init__()
        error_msg: Optional[str] = None

        if isinstance(self.describe_method, str):
            self.describe_method = [self.describe_method]

        invalid_methods: list[str] = [
            method
            for method in self.describe_method
            if method not in self.AVAILABLE_DESCRIBE_METHODS
        ]
        if invalid_methods:
            error_msg = (
                f"Invalid describe method: {invalid_methods}. "
                f"Available methods are: {self.AVAILABLE_DESCRIBE_METHODS}"
            )
            self._logger.error(error_msg)
            raise UnsupportedMethodError(error_msg)

        if "percentile" not in self.describe_method:
            self.percentile = None
            self._logger.debug(
                "Percentile method is not in describe_method, "
                "setting percentile to None"
            )

    def update_data(self, data: dict[str, pd.DataFrame]) -> None:
        error_msg: Optional[str] = None

        self._logger.info(
            f"Updating data with {len(self.REQUIRED_INPUT_KEYS)} required keys"
        )

        # Validate required keys
        if not all(key in data for key in self.REQUIRED_INPUT_KEYS):
            error_msg = f"Missing required keys. Expected: {self.REQUIRED_INPUT_KEYS}"
            self._logger.error(error_msg)
            raise ConfigError(error_msg)


class DataDescriber(BaseEvaluator):
    """
    Evaluator for Desciber method.
    """

    REQUIRED_INPUT_KEYS: list[str] = ["data"]
    EXEC_GRANULARITY_MAP: dict[str, str] = {
        method: granularity
        for granularity, methods in {
            "global": [
                "row_count",
                "col_count",
                "global_na_count",
            ],
            "columnwise": [
                "mean",
                "median",
                "std",
                "var",
                "min",
                "max",
                "kurtosis",
                "skew",
                "q1",
                "q3",
                "iqr",
                "range",
                "percentile",
                "nunique",
                "col_na_count",
            ],
            "pairwise": [
                "cov",
                "corr",
            ],
        }.items()
        for method in methods
    }
    MODULE_MAP: dict[str, callable] = {
        "row_count": DataDescriberRowCount,
        "col_count": DataDescriberColumnCount,
        "global_na_count": DataDescriberGlobalNA,
        "mean": DataDescriberMean,
        "median": DataDescriberMedian,
        "std": DataDescriberStd,
        "var": DataDescriberVar,
        "min": DataDescriberMin,
        "max": DataDescriberMax,
        "kurtosis": DataDescriberKurtosis,
        "skew": DataDescriberSkew,
        "q1": DataDescriberQ1,
        "q3": DataDescriberQ3,
        "percentile": DataDescriberPercentile,
        "col_na_count": DataDescriberColNA,
        "nunique": DataDescriberNUnique,
        "corr": DataDescriberCorr,
        "cov": DataDescriberCov,
    }
    INT_MODULE: list[str] = [
        "row_count",
        "col_count",
        "global_na_count",
        "col_na_count",
        "nunique",
    ]
    AVAILABLE_SCORES_GRANULARITY: list[str] = ["global", "columnwise", "pairwise"]

    def __init__(self, config: dict):
        """
        Args:
            config (dict): The configuration assign by Evaluator.

        Attributes:
            REQUIRED_INPUT_KEYS (list[str]): The required input keys.
            AVAILABLE_SCORES_GRANULARITY (list[str]): The available scores granularity.
            _logger (logging.Logger): The logger object.
            desc_config (StatsConfig): The configuration parameters for the describing evaluator.
            _impl (Optional[dict[str, callable]]): The implementation object.
        """
        super().__init__(config=config)
        self._logger: logging.Logger = logging.getLogger(
            f"PETsARD.{self.__class__.__name__}"
        )

        self._logger.debug(f"Verifying StatsConfig with parameters: {self.config}")
        self.desc_config = DataDescriberConfig(**self.config)
        self._logger.debug("StatsConfig successfully initialized")

        impl_config: dict[str, Any] = self.config
        _ = impl_config.pop("eval_method", None)
        self._impl: Optional[dict[str, callable]] = {
            method: self.MODULE_MAP[method](config=impl_config)
            for method in self.desc_config.describe_method
        }

    def _get_columnwise(self, columnwise_desc: list[Any]) -> pd.DataFrame:
        columnwise_desc_df: pd.DataFrame = pd.DataFrame(
            {k: v for d in columnwise_desc for k, v in d.items()}
        )

        for col in columnwise_desc_df.columns:
            if col in self.INT_MODULE:
                columnwise_desc_df[col] = (
                    columnwise_desc_df[col].fillna(-1).astype(int).replace(-1, pd.NA)
                )
            else:
                # 安全處理 safe_round 可能返回 None 的情況
                rounded_series = columnwise_desc_df[col].apply(safe_round)
                columnwise_desc_df[col] = rounded_series.fillna(pd.NA)

        return columnwise_desc_df

    def _eval(self, data: dict[str, pd.DataFrame]) -> dict[str, pd.DataFrame]:
        """
        Evaluating the describer.
            _impl should be initialized in this method.

        Args:
            data (dict[str, pd.DataFrame]): The data to be described.

        Return:
            (dict[str, pd.DataFrame]): The describe result
        """
        error_msg: str = None
        self._logger.info(
            f"Starting evaluation with {len(self.desc_config.describe_method)} statistical methods"
        )
        self.desc_config.update_data(data)

        # Assign to list first, convert to DataFrame later
        temp_desc_result: dict[str, list[Any]] = {}
        granularity: str = None
        for method in self.desc_config.describe_method:
            self._logger.info(f"Describing data with method: {method}")

            try:
                granularity = self.EXEC_GRANULARITY_MAP[method]
            except KeyError:
                error_msg: str = f"Unassigned granulariy describer method: {method}"
                self._logger.error(error_msg)
                raise UnsupportedMethodError(error_msg)
            if granularity not in temp_desc_result:
                temp_desc_result[granularity] = []

            temp_desc_result[granularity].append(self._impl[method].eval(data["data"]))

        desc_result: dict[str, pd.DataFrame] = {}
        if "global" in temp_desc_result:
            desc_result["global"] = pd.DataFrame(
                [{k: v for d in temp_desc_result["global"] for k, v in d.items()}]
            )
        if "columnwise" in temp_desc_result:
            desc_result["columnwise"] = self._get_columnwise(
                temp_desc_result["columnwise"]
            )
        if "pairwise" in temp_desc_result:
            temp_df: pd.DataFrame = pd.DataFrame(
                [
                    (col1, col2, metric, value)
                    for d in temp_desc_result["pairwise"]
                    for metric, pairs in d.items()
                    for (col1, col2), value in pairs.items()
                ],
                columns=["column1", "column2", "metric", "value"],
            ).reset_index(drop=True)

            temp_df = temp_df.pivot(
                index=["column1", "column2"], columns="metric", values="value"
            ).reset_index()
            temp_df.columns.name = None
            desc_result["pairwise"] = temp_df.copy()

        return desc_result



from enum import Enum


class DataType(Enum):
    """
    Standard data types for metadata system

    Attr.:
        BOOLEAN (str): Boolean type
        INT8 (str): 8-bit integer
        INT16 (str): 16-bit integer
        INT32 (str): 32-bit integer
        INT64 (str): 64-bit integer
        FLOAT32 (str): 32-bit floating point
        FLOAT64 (str): 64-bit floating point
        DECIMAL (str): Decimal type
        STRING (str): String type
        BINARY (str): Binary type
        DATE (str): Date type
        TIME (str): Time type
        TIMESTAMP (str): Timestamp type
        TIMESTAMP_TZ (str): Timestamp with timezone type
        UUID (str): UUID type
        FIXED (str): Fixed-length byte array type
    """

    # Primitive types
    BOOLEAN = "boolean"
    INT8 = "int8"
    INT16 = "int16"
    INT32 = "int32"
    INT64 = "int64"
    FLOAT32 = "float32"
    FLOAT64 = "float64"
    DECIMAL = "decimal"
    STRING = "string"
    BINARY = "binary"

    # Date/Time types
    DATE = "date"
    TIME = "time"
    TIMESTAMP = "timestamp"
    TIMESTAMP_TZ = "timestamp_tz"

    # Special types
    UUID = "uuid"  # Universally Unique Identifier
    FIXED = "fixed"  # Fixed-length byte array


class LogicalType(Enum):
    """
    Logical types that provide semantic meaning

    Attr.:
        EMAIL (str): Email address
        URL (str): URL
        JSON (str): JSON object
        PHONE (str): Phone number
        CURRENCY (str): Currency value
        PERCENTAGE (str): Percentage value
        CATEGORICAL (str): Categorical data
        ORDINAL (str): Ordinal data
        LATITUDE (str): Latitude coordinate
        LONGITUDE (str): Longitude coordinate
        EMBEDDING (str): Vector embedding
        IP_ADDRESS (str): IP address (IPv4)
        IPV6_ADDRESS (str): IPv6 address
        MAC_ADDRESS (str): MAC address
        UUID (str): UUID
        CREDIT_CARD (str): Credit card number (masked detection only)
        POSTAL_CODE (str): Postal/ZIP code
        COLOR_HEX (str): Hexadecimal color code
        HASH_MD5 (str): MD5 hash
        HASH_SHA256 (str): SHA256 hash
        FILE_PATH (str): File system path
        DATETIME_ISO (str): ISO 8601 datetime string
    """

    EMAIL = "email"
    URL = "url"
    JSON = "json"
    PHONE = "phone"
    CURRENCY = "currency"
    PERCENTAGE = "percentage"
    CATEGORICAL = "categorical"
    ORDINAL = "ordinal"
    LATITUDE = "latitude"
    LONGITUDE = "longitude"
    EMBEDDING = "embedding"
    IP_ADDRESS = "ip_address"
    IPV6_ADDRESS = "ipv6_address"
    MAC_ADDRESS = "mac_address"
    UUID = "uuid"
    CREDIT_CARD = "credit_card"
    POSTAL_CODE = "postal_code"
    COLOR_HEX = "color_hex"
    HASH_MD5 = "hash_md5"
    HASH_SHA256 = "hash_sha256"
    FILE_PATH = "file_path"
    DATETIME_ISO = "datetime_iso"


def legacy_safe_round(value: float, decimals: int = 4) -> float:
    """Safe round function"""
    return round(value, decimals)



import unittest
from unittest.mock import MagicMock, patch

import pandas as pd

from petsard.evaluator.customer_evaluator import CustomEvaluator


class TestCustomEvaluator(unittest.TestCase):
    """Test for Custom Evaluator."""

    def setUp(self):
        """Set up test fixtures."""
        # Test configuration
        self.config = {
            "eval_method": "custom_method",
            "module_path": "/path/to/custom_module.py",
            "class_name": "MyCustomEvaluator",
        }

        # Create mock data
        self.data = {
            "ori": pd.DataFrame(
                {
                    "col1": [1, 2, 3, 4, 5],
                    "col2": ["a", "b", "c", "d", "e"],
                    "col3": [1.1, 2.2, 3.3, 4.4, 5.5],
                }
            ),
            "syn": pd.DataFrame(
                {
                    "col1": [1, 2, 3, 4, 5],
                    "col2": ["a", "b", "c", "d", "e"],
                    "col3": [1.1, 2.2, 3.3, 4.4, 5.5],
                }
            ),
        }

    @patch("petsard.util.load_external_module")
    def test_init(self, mock_load_module):
        """Test initialization."""
        # Create mock custom evaluator class
        mock_evaluator_class = MagicMock()
        mock_evaluator_class.REQUIRED_INPUT_KEYS = ["ori", "syn"]
        mock_evaluator_class.AVAILABLE_SCORES_GRANULARITY = ["global"]

        # Setup mock return for the load_external_module
        mock_load_module.return_value = (None, mock_evaluator_class)

        # Initialize the CustomEvaluator
        evaluator = CustomEvaluator(config=self.config)

        # Verify load_external_module was called with correct arguments
        mock_load_module.assert_called_once_with(
            module_path="/path/to/custom_module.py",
            class_name="MyCustomEvaluator",
            logger=evaluator._logger,
            required_methods=CustomEvaluator.REQUIRED_METHODS,
        )

        # Verify properties were set correctly
        self.assertEqual(evaluator.config, self.config)
        self.assertEqual(evaluator.REQUIRED_INPUT_KEYS, ["ori", "syn"])
        self.assertEqual(evaluator.AVAILABLE_SCORES_GRANULARITY, ["global"])

        # Verify the mock custom evaluator class was instantiated
        mock_evaluator_class.assert_called_once_with(config=self.config)

    @patch("petsard.util.load_external_module")
    def test_eval(self, mock_load_module):
        """Test eval method."""
        # Create mock custom evaluator instance
        mock_evaluator_instance = MagicMock()
        expected_result = {
            "global": pd.DataFrame({"score": [0.85]}),
            "details": {"metric1": 0.8, "metric2": 0.9},
        }
        mock_evaluator_instance.eval.return_value = expected_result

        # Create mock custom evaluator class
        mock_evaluator_class = MagicMock(return_value=mock_evaluator_instance)
        mock_evaluator_class.REQUIRED_INPUT_KEYS = ["ori", "syn"]
        mock_evaluator_class.AVAILABLE_SCORES_GRANULARITY = ["global", "details"]

        # Setup mock return
        mock_load_module.return_value = (None, mock_evaluator_class)

        # Initialize and evaluate
        evaluator = CustomEvaluator(config=self.config)
        result = evaluator._eval(self.data)

        # Verify the eval method was called on the custom evaluator
        mock_evaluator_instance.eval.assert_called_once_with(data=self.data)

        # Verify the results match
        self.assertEqual(result, expected_result)

    def test_missing_module_path(self):
        """Test initialization with missing module_path."""
        invalid_config = self



from pathlib import Path
from unittest.mock import MagicMock, patch

import pandas as pd
import pytest

from petsard.exceptions import ConfigError, UnsupportedMethodError
from petsard.loader.loader import Loader, LoaderConfig, LoaderFileExt


class TestLoaderConfig:
    """Test cases for LoaderConfig class
    LoaderConfig 類的測試案例
    """

    def test_config_requires_filepath_or_method(self):
        """Test that either filepath or method must be specified
        測試必須指定 filepath 或 method 參數
        """
        with pytest.raises(ConfigError):
            LoaderConfig()

    def test_default_method(self):
        """Test default method configuration
        測試默認方法配置
        """
        with patch.object(LoaderConfig, "_load_benchmark_config") as mock_load_config:
            mock_load_config.return_value = {
                "adult-income": {
                    "filename": "adult-income.csv",
                    "access": "public",
                    "region_name": "us-west-2",
                    "bucket_name": "petsard-benchmark",
                    "sha256": "1f13ee2bf9d7c66098429281ab91fa1b51cbabd3b805cc365b3c6b44491ea2c0",
                }
            }

            config = LoaderConfig(method="default")
            # 檢查初始 filepath 被設置為 benchmark URL
            # Check that initial filepath is set to benchmark URL
            assert config.DEFAULT_METHOD_FILEPATH == "benchmark://adult-income"
            # 檢查 filepath 已被處理為本地路徑
            # Check that filepath has been processed to local path
            assert str(config.filepath).endswith("benchmark/adult-income.csv")
            # 檢查 benchmark 標記已被設置
            # Check that benchmark flag is set
            assert config.benchmark
            assert config.benchmark_name == "adult-income"

    def test_unsupported_method(self):
        """Test unsupported method raises error
        測試不支援的方法會引發錯誤
        """
        with pytest.raises(UnsupportedMethodError):
            LoaderConfig(method="unsupported_method")

    def test_benchmark_path_parsing(self):
        """Test parsing of benchmark path
        測試基準資料集路徑解析
        """
        with patch.object(LoaderConfig, "_load_benchmark_config") as mock_load_config:
            mock_load_config.return_value = {
                "adult-income": {
                    "filename": "adult.csv",
                    "access": "public",
                    "region_name": "us-west-2",
                    "bucket_name": "test-bucket",
                    "sha256": "test-hash",
                }
            }
            config = LoaderConfig(filepath="benchmark://adult-income")
            assert config.benchmark
            assert config.benchmark_name == "adult-income"
            assert config.filepath == Path("benchmark").joinpath("adult.csv")
            assert config.benchmark_filename == "adult.csv"
            assert config.benchmark_access == "public"
            assert config.benchmark_region_name == "us-west-2"
            assert config.benchmark_bucket_name == "test-bucket"
            assert config.benchmark_sha256 == "test-hash"

    def test_unsupported_benchmark(self):
        """Test unsupported benchmark raises error
        測試不支援的基準資料集會引發錯誤
        """
        with patch.object(LoaderConfig, "_load_benchmark_config") as mock_load_config:
            mock_load_config.return_value = {}
            with pytest.raises(UnsupportedMethodError):
                LoaderConfig(filepath="benchmark://nonexistent")

    def test_private_benchmark_unsupported(self):
        """Test private benchmark access raises error
        測試私有基準資料集存取會引發錯誤
        """
        with patch.object(LoaderConfig, "_load_benchmark_config") as mock_load_config:
            mock_load_config.return_value = {
                "private-data": {
                    "filename": "private.csv",
                    "access": "private",
                    "region_name": "us-west-2",
                    "bucket_name": "private-bucket",
                    "sha256": "test-hash",
                }
            }
            with pytest.raises(UnsupportedMethodError):
                LoaderConfig(filepath="benchmark://private-data")

    @pytest.mark.parametrize(
        "filepath,expected_ext,expected_code",
        [
            ("path/to/file.csv", ".csv", LoaderFileExt.CSVTYPE),
            ("path/to/file.xlsx", ".xlsx", LoaderFileExt.EXCELTYPE),
            ("path/to/file.xls", ".xls", LoaderFileExt.EXCELTYPE),
            ("path/to/file.CSV", ".csv", LoaderFileExt.CSVTYPE),
            ("path/to/file.XLSX", ".xlsx", LoaderFileExt.EXCELTYPE),
        ],
    )
    def test_file_extension_handling(self, filepath, expected_ext, expected_code):
        """Test file extension parsing and mapping
        測試檔案副檔名解析和映射
        """
        config = LoaderConfig(filepath=filepath)
        assert config.file_ext == expected_ext
        assert config.file_ext_code == expected_code

    def test_invalid_file_extension(self):
        """Test handling of invalid file extensions
        測試處理無效的檔案副檔名
        """
        with pytest.raises(UnsupportedMethodError):
            LoaderConfig(filepath="path/to/file.invalid")

    def test_unsupported_column_type(self):
        """Test handling of unsupported column types
        測試處理不支援的欄位類型
        """
        with pytest.raises(UnsupportedMethodError):
            LoaderConfig(
                filepath="path/to/file.csv", column_types={"unsupported_type": ["col1"]}
            )

    def test_get_method(self):
        """Test get() method returns config dictionary
        測試 get() 方法返回配置字典
        """
        config = LoaderConfig(filepath="path/to/file.csv")
        config_dict = config.get()
        assert isinstance(config_dict, dict)
        assert config_dict["filepath"] == "path/to/file.csv"
        assert config_dict["file_ext"] == ".csv"


class TestLoader:
    """Test cases for main Loader functionality
    主要 Loader 功能的測試案例
    """

    @pytest.fixture
    def sample_csv_path(self, tmp_path):
        """Create a temporary CSV file for testing
        創建臨時 CSV 檔案用於測試
        """
        csv_file = tmp_path / "test.csv"
        pd.DataFrame({"A": [1, 2, 3], "B": ["x", "y", "z"]}).to_csv(
            csv_file, index=False
        )
        return str(csv_file)

    def test_loader_init_no_config(self):
        """Test Loader initialization with no config
        測試沒有配置的 Loader 初始化
        """
        with pytest.raises(ConfigError):
            Loader()

    @pytest.mark.parametrize(
        "filepath,expected_ext",
        [
            ("path/to/file.csv", ".csv"),
            ("path.with.dots/file.csv", ".csv"),
            ("path/to/file.name.with.dots.csv", ".csv"),
            ("./relative/path/file.csv", ".csv"),
            ("../parent/path/file.csv", ".csv"),
            ("/absolute/path/file.csv", ".csv"),
            ("file.CSV", ".csv"),  # 測試大小寫 / Test case sensitivity
            ("path/to/file.XLSX", ".xlsx"),
        ],
    )
    def test_handle_filepath_with_complex_name(self, filepath, expected_ext):
        """Test handling of complex file paths
        測試處理複雜的檔案路徑
        > issue 375
        """
        loader = Loader(filepath=filepath)
        assert loader.config.file_ext == expected_ext
        assert loader.config.filepath == filepath

    def test_loader_init_with_filepath(self, sample_csv_path):
        """Test Loader initialization with filepath
        測試使用檔案路徑初始化 Loader
        """
        loader = Loader(filepath=sample_csv_path)
        assert loader.config.filepath == sample_csv_path
        assert loader.config.file_ext == ".csv"

    def test_loader_init_with_column_types(self, sample_csv_path):
        """Test Loader initialization with column types
        測試使用欄位類型初始化 Loader
        """
        column_types = {"category": ["B"]}
        loader = Loader(filepath=sample_csv_path, column_types=column_types)
        assert loader.config.column_types == column_types

    def test_benchmark_loader(self):
        """Test loading benchmark dataset
        測試載入基準資料集
        """
        with (
            patch("petsard.loader.loader.BenchmarkerRequests") as mock_benchmarker,
            patch.object(LoaderConfig, "_load_benchmark_config") as mock_load_config,
        ):
            mock_load_config.return_value = {
                "adult-income": {
                    "filename": "adult.csv",
                    "access": "public",
                    "region_name": "us-west-2",
                    "bucket_name": "test-bucket",
                    "sha256": "test-hash",
                }
            }

            loader = Loader(filepath="benchmark://adult-income")
            # Benchmarker should not be called during init
            # 初始化期間不應調用 Benchmarker
            mock_benchmarker.assert_not_called()

            assert loader.config.benchmark
            assert loader.config.benchmark_name == "adult-income"

    def test_load_csv(self, sample_csv_path):
        """Test loading CSV file
        測試載入 CSV 檔案
        """
        loader = Loader(filepath=sample_csv_path)

        with (
            patch("pandas.read_csv") as mock_read_csv,
            patch(
                "petsard.metadater.metadater.Metadater.create_schema"
            ) as mock_create_schema,
            patch(
                "petsard.metadater.schema.schema_functions.apply_schema_transformations"
            ) as mock_apply_transformations,
        ):
            # Setup mock returns
            # 設置模擬回傳值
            mock_df = pd.DataFrame({"A": [1, 2, 3], "B": ["x", "y", "z"]})
            mock_read_csv.return_value = mock_df.fillna(pd.NA)

            mock_schema_instance = MagicMock()
            mock_create_schema.return_value = mock_schema_instance
            mock_apply_transformations.return_value = mock_df

            # Call load method
            # 調用 load 方法
            data, metadata = loader.load()

            # Assertions
            # 斷言
            mock_read_csv.assert_called_once_with(
                sample_csv_path,
                header="infer",
                names=None,
                na_values=loader.config.na_values,
                dtype=None,
            )
            mock_create_schema.assert_called_once()
            mock_apply_transformations.assert_called_once()
            assert data is not None
            assert metadata is not None

    def test_load_excel(self):
        """Test loading Excel file
        測試載入 Excel 檔案
        """
        excel_path = "path/to/file.xlsx"
        loader = Loader(filepath=excel_path)

        with (
            patch("pandas.read_excel") as mock_read_excel,
            patch(
                "petsard.metadater.metadater.Metadater.create_schema"
            ) as mock_create_schema,
            patch(
                "petsard.metadater.schema.schema_functions.apply_schema_transformations"
            ) as mock_apply_transformations,
            patch("os.path.exists") as mock_exists,
        ):
            # Setup mock returns
            # 設置模擬回傳值
            mock_exists.return_value = True
            mock_df = pd.DataFrame({"A": [1, 2, 3], "B": ["x", "y", "z"]})
            mock_read_excel.return_value = mock_df.fillna(pd.NA)

            mock_schema_instance = MagicMock()
            mock_create_schema.return_value = mock_schema_instance
            mock_apply_transformations.return_value = mock_df

            # Call load method
            # 調用 load 方法
            data, metadata = loader.load()

            # Assertions
            # 斷言
            mock_read_excel.assert_called_once_with(
                excel_path,
                header="infer",
                names=None,
                na_values=loader.config.na_values,
                dtype=None,
            )
            mock_create_schema.assert_called_once()
            mock_apply_transformations.assert_called_once()
            assert data is not None
            assert metadata is not None

    def test_benchmark_data_load(self):
        """Test loading benchmark data
        測試載入基準資料
        """
        with (
            patch.object(LoaderConfig, "_load_benchmark_config") as mock_load_config,
            patch("petsard.loader.loader.BenchmarkerRequests") as mock_benchmarker,
            patch("pandas.read_csv") as mock_read_csv,
            patch(
                "petsard.metadater.metadater.Metadater.create_schema"
            ) as mock_create_schema,
            patch(
                "petsard.metadater.schema.schema_functions.apply_schema_transformations"
            ) as mock_apply_transformations,
        ):
            # Setup mock returns
            # 設置模擬回傳值
            mock_load_config.return_value = {
                "adult-income": {
                    "filename": "adult.csv",
                    "access": "public",
                    "region_name": "us-west-2",
                    "bucket_name": "test-bucket",
                    "sha256": "test-hash",
                }
            }
            mock_df = pd.DataFrame({"A": [1, 2, 3], "B": ["x", "y", "z"]})
            mock_read_csv.return_value = mock_df.fillna(pd.NA)

            mock_schema_instance = MagicMock()
            mock_create_schema.return_value = mock_schema_instance
            mock_apply_transformations.return_value = mock_df
            mock_benchmarker_instance = MagicMock()
            mock_benchmarker.return_value = mock_benchmarker_instance

            # Create and load benchmark data
            # 創建和載入基準資料
            loader = Loader(filepath="benchmark://adult-income")
            data, metadata = loader.load()

            # Assertions
            # 斷言
            mock_benchmarker.assert_called_once()
            mock_benchmarker_instance.download.assert_called_once()
            mock_read_csv.assert_called_once()
            mock_create_schema.assert_called_once()
            mock_apply_transformations.assert_called_once()
            assert data is not None
            assert metadata is not None

    def test_custom_na_values(self, sample_csv_path):
        """Test loading with custom NA values
        測試使用自定義 NA 值載入資料
        """
        na_values = ["x"]
        loader = Loader(filepath=sample_csv_path, na_values=na_values)

        with (
            patch("pandas.read_csv") as mock_read_csv,
            patch(
                "petsard.metadater.metadater.Metadater.create_schema"
            ) as mock_create_schema,
            patch(
                "petsard.metadater.schema.schema_functions.apply_schema_transformations"
            ) as mock_apply_transformations,
        ):
            # Setup mock returns
            # 設置模擬回傳值
            mock_df = pd.DataFrame({"A": [1, 2, 3], "B": [None, "y", "z"]})
            mock_read_csv.return_value = mock_df.fillna(pd.NA)

            mock_schema_instance = MagicMock()
            mock_create_schema.return_value = mock_schema_instance
            mock_apply_transformations.return_value = mock_df

            # Call load method
            # 調用 load 方法
            data, metadata = loader.load()

            # Assertions
            # 斷言
            mock_read_csv.assert_called_once_with(
                sample_csv_path,
                header="infer",
                names=None,
                na_values=na_values,
                dtype=None,
            )
            mock_create_schema.assert_called_once()
            mock_apply_transformations.assert_called_once()
            assert data is not None
            assert metadata is not None

    def test_custom_header_names(self, sample_csv_path):
        """Test loading with custom header names
        測試使用自定義欄位名稱載入資料
        """
        header_names = ["Col1", "Col2"]
        loader = Loader(filepath=sample_csv_path, header_names=header_names)

        with (
            patch("pandas.read_csv") as mock_read_csv,
            patch(
                "petsard.metadater.metadater.Metadater.create_schema"
            ) as mock_create_schema,
            patch(
                "petsard.metadater.schema.schema_functions.apply_schema_transformations"
            ) as mock_apply_transformations,
        ):
            # Setup mock returns
            # 設置模擬回傳值
            mock_df = pd.DataFrame({"Col1": [1, 2, 3], "Col2": ["x", "y", "z"]})
            mock_read_csv.return_value = mock_df.fillna(pd.NA)

            mock_schema_instance = MagicMock()
            mock_create_schema.return_value = mock_schema_instance
            mock_apply_transformations.return_value = mock_df

            # Call load method
            # 調用 load 方法
            data, metadata = loader.load()

            # Assertions
            # 斷言
            mock_read_csv.assert_called_once_with(
                sample_csv_path,
                header=0,
                names=header_names,
                na_values=loader.config.na_values,
                dtype=None,
            )
            mock_create_schema.assert_called_once()
            mock_apply_transformations.assert_called_once()
            assert data is not None
            assert metadata is not None


class TestLoaderAmbiguousDataFeatures:
    """Test cases for ambiguous data type processing features
    容易誤判、型別判斷模糊資料處理功能的測試案例
    """

    @pytest.fixture
    def ambiguous_sample_csv(self, tmp_path):
        """Create a sample CSV file with ambiguous data types for testing
        創建含有型別判斷模糊資料的測試用 CSV 檔案
        """
        csv_file = tmp_path / "ambiguous_test.csv"
        # Ambiguous data with leading zeros, mixed types, and null values
        test_data = {
            "id_code": ["001", "002", "010", "099"],  # Leading zeros - 前導零代號
            "age": [25, 30, "", 45],  # Integers with null - 含空值整數
            "amount": [
                50000.5,
                60000.7,
                "",
                75000.3,
            ],  # Floats with null - 含空值浮點數
            "score": [85, 90, 78, 92],  # Pure integers - 純整數
            "category": ["A級", "B級", "C級", "A級"],  # Categories - 分類資料
        }
        pd.DataFrame(test_data).to_csv(csv_file, index=False)
        return str(csv_file)

    def test_preserve_raw_data_feature(self, ambiguous_sample_csv):
        """Test preserve_raw_data feature prevents automatic type inference
        測試 preserve_raw_data 功能阻止自動類型推斷
        """
        loader = Loader(
            filepath=ambiguous_sample_csv,
            preserve_raw_data=True,
            auto_detect_leading_zeros=True,
            force_nullable_integers=True,
        )

        with (
            patch("pandas.read_csv") as mock_read_csv,
            patch(
                "petsard.metadater.metadater.Metadater.create_schema"
            ) as mock_create_schema,
            patch(
                "petsard.metadater.schema.schema_functions.apply_schema_transformations"
            ) as mock_apply_transformations,
        ):
            # Setup mock returns
            mock_df = pd.DataFrame(
                {
                    "id_code": ["001", "002", "010", "099"],
                    "age": ["25", "30", "", "45"],
                    "amount": ["50000.5", "60000.7", "", "75000.3"],
                    "score": ["85", "90", "78", "92"],
                    "category": ["A級", "B級", "C級", "A級"],
                }
            )
            mock_read_csv.return_value = mock_df.fillna(pd.NA)

            mock_schema_instance = MagicMock()
            mock_create_schema.return_value = mock_schema_instance
            mock_apply_transformations.return_value = mock_df

            # Call load method
            data, metadata = loader.load()

            # Verify that dtype="object" was used to preserve raw data
            mock_read_csv.assert_called_once_with(
                ambiguous_sample_csv,
                header="infer",
                names=None,
                na_values=loader.config.na_values,
                dtype="object",  # This should be "object" string when preserve_raw_data=True
            )
            mock_create_schema.assert_called_once()
            mock_apply_transformations.assert_called_once()
            assert data is not None
            assert metadata is not None

    def test_leading_zero_detection_config(self, ambiguous_sample_csv):
        """Test auto_detect_leading_zeros configuration
        測試 auto_detect_leading_zeros 配置
        """
        # Test with leading zero detection enabled
        loader_on = Loader(
            filepath=ambiguous_sample_csv,
            auto_detect_leading_zeros=True,
        )
        assert loader_on.config.auto_detect_leading_zeros is True

        # Test with leading zero detection disabled
        loader_off = Loader(
            filepath=ambiguous_sample_csv,
            auto_detect_leading_zeros=False,
        )
        assert loader_off.config.auto_detect_leading_zeros is False

    def test_nullable_integer_config(self, ambiguous_sample_csv):
        """Test force_nullable_integers configuration
        測試 force_nullable_integers 配置
        """
        # Test with nullable integers enabled
        loader_on = Loader(
            filepath=ambiguous_sample_csv,
            force_nullable_integers=True,
        )
        assert loader_on.config.force_nullable_integers is True

        # Test with nullable integers disabled
        loader_off = Loader(
            filepath=ambiguous_sample_csv,
            force_nullable_integers=False,
        )
        assert loader_off.config.force_nullable_integers is False

    def test_ambiguous_data_config_combination(self, ambiguous_sample_csv):
        """Test combination of ambiguous data processing configurations
        測試容易誤判資料處理配置的組合
        """
        loader = Loader(
            filepath=ambiguous_sample_csv,
            preserve_raw_data=True,
            auto_detect_leading_zeros=True,
            force_nullable_integers=True,
        )

        # Verify all configurations are set correctly
        assert loader.config.preserve_raw_data is True
        assert loader.config.auto_detect_leading_zeros is True
        assert loader.config.force_nullable_integers is True

    def test_backward_compatibility(self, ambiguous_sample_csv):
        """Test that new features don't break existing functionality
        測試新功能不會破壞現有功能
        """
        # Test with default settings (all new features disabled)
        loader = Loader(filepath=ambiguous_sample_csv)

        # Verify default values
        assert loader.config.preserve_raw_data is False
        assert loader.config.auto_detect_leading_zeros is False
        assert loader.config.force_nullable_integers is False

        with (
            patch("pandas.read_csv") as mock_read_csv,
            patch(
                "petsard.metadater.metadater.Metadater.create_schema"
            ) as mock_create_schema,
            patch(
                "petsard.metadater.schema.schema_functions.apply_schema_transformations"
            ) as mock_apply_transformations,
        ):
            # Setup mock returns
            mock_df = pd.DataFrame({"A": [1, 2, 3], "B": ["x", "y", "z"]})
            mock_read_csv.return_value = mock_df.fillna(pd.NA)

            mock_schema_instance = MagicMock()
            mock_create_schema.return_value = mock_schema_instance
            mock_apply_transformations.return_value = mock_df

            # Call load method
            data, metadata = loader.load()

            # Verify normal behavior (no dtype parameter when preserve_raw_data=False)
            mock_read_csv.assert_called_once_with(
                ambiguous_sample_csv,
                header="infer",
                names=None,
                na_values=loader.config.na_values,
                # No dtype parameter should be passed when preserve_raw_data=False
            )
            assert data is not None
            assert metadata is not None


class TestLoaderFileExt:
    """Test cases for LoaderFileExt class
    LoaderFileExt 類的測試案例
    """

    @pytest.mark.parametrize(
        "file_ext,expected_code",
        [
            (".csv", LoaderFileExt.CSVTYPE),
            (".CSV", LoaderFileExt.CSVTYPE),
            (".xlsx", LoaderFileExt.EXCELTYPE),
            (".XLSX", LoaderFileExt.EXCELTYPE),
            (".xls", LoaderFileExt.EXCELTYPE),
            (".xlsm", LoaderFileExt.EXCELTYPE),
            (".xlsb", LoaderFileExt.EXCELTYPE),
            (".ods", LoaderFileExt.EXCELTYPE),
            (".odt", LoaderFileExt.EXCELTYPE),
        ],
    )
    def test_get_file_ext_code(self, file_ext, expected_code):
        """Test getting file extension code
        測試獲取檔案副檔名類型
        """
        assert LoaderFileExt.get(file_ext) == expected_code

    def test_unsupported_file_ext(self):
        """Test handling of unsupported file extensions
        測試處理不支援的檔案副檔名
        """
        with pytest.raises(KeyError):
            LoaderFileExt.get(".unsupported")



import numpy as np
import pandas as pd
import pytest
from sklearn.preprocessing import StandardScaler

from petsard.exceptions import UnfittedError
from petsard.processor.scaler import (
    ScalerLog,
    ScalerMinMax,
    ScalerStandard,
    ScalerTimeAnchor,
    ScalerZeroCenter,
)


class BaseScalerTest:
    """Base class for all scaler tests"""

    @property
    def scaler_class(self):
        """The scaler class to be tested"""
        raise NotImplementedError

    def test_unfitted_error(self):
        """Test that unfitted scaler raises proper error"""
        df_data = pd.DataFrame({"col1": [1.0, 2.0, 3.0]})
        scaler = self.scaler_class()

        with pytest.raises(UnfittedError):
            scaler.transform(df_data["col1"])

        with pytest.raises(UnfittedError):
            scaler.inverse_transform(df_data["col1"])


class Test_ScalerStandard(BaseScalerTest):
    @property
    def scaler_class(self):
        return ScalerStandard

    def test_standard_scaling(self):
        # Create a sample dataframe
        df_data = pd.DataFrame({"col1": [1.0, 2.0, 3.0], "col2": [4.0, 5.0, 6.0]})
        df_expected = pd.DataFrame(
            StandardScaler().fit_transform(df_data), columns=["col1", "col2"]
        )

        # Create an instance of the class
        scaler = self.scaler_class()
        scaler.fit(df_data["col1"])
        transformed = scaler.transform(df_data["col1"])

        # Assert that the dataframe is correct
        assert list(transformed) == list(df_expected["col1"].values)


class Test_ScalerZeroCenter(BaseScalerTest):
    @property
    def scaler_class(self):
        return ScalerZeroCenter

    def test_zero_center_scaling(self):
        # Create a sample dataframe
        df_data = pd.DataFrame({"col1": [1.0, 2.0, 3.0], "col2": [4.0, 5.0, 6.0]})
        df_expected = pd.DataFrame({"col1": [-1.0, 0.0, 1.0], "col2": [-1.0, 0.0, 1.0]})

        # Create an instance of the class
        scaler = self.scaler_class()
        scaler.fit(df_data["col1"])
        transformed = scaler.transform(df_data["col1"])

        # Assert that the dataframe is correct
        assert list(transformed) == list(df_expected["col1"].values)


class Test_ScalerMinMax(BaseScalerTest):
    @property
    def scaler_class(self):
        return ScalerMinMax

    def test_minmax_scaling(self):
        # Create a sample dataframe
        df_data = pd.DataFrame({"col1": [1.0, 2.0, 3.0], "col2": [4.0, 5.0, 6.0]})
        df_expected = pd.DataFrame({"col1": [0.0, 0.5, 1.0], "col2": [0.0, 0.5, 1.0]})

        # Create an instance of the class
        scaler = self.scaler_class()
        scaler.fit(df_data["col1"])
        transformed = scaler.transform(df_data["col1"])

        # Assert that the dataframe is correct
        assert list(transformed) == list(df_expected["col1"].values)


class Test_ScalerLog(BaseScalerTest):
    @property
    def scaler_class(self):
        return ScalerLog

    def test_log_scaling(self):
        # Create a sample dataframe with positive values
        df_data = pd.DataFrame({"col1": [1.0, 2.0, 3.0]})

        # Create an instance of the class
        scaler = self.scaler_class()
        scaler.fit(df_data["col1"])
        transformed = scaler.transform(df_data["col1"])

        # Test expected log values
        expected = np.log(df_data["col1"].values)
        assert np.allclose(transformed.ravel(), expected)

    def test_negative_values(self):
        # Test negative values
        df_negative = pd.DataFrame({"col1": [-1.0, 0.0, 1.0]})
        scaler = self.scaler_class()
        with pytest.raises(ValueError):
            scaler.fit(df_negative["col1"])


class Test_ScalerTimeAnchor(BaseScalerTest):
    @property
    def scaler_class(self):
        return ScalerTimeAnchor

    def test_time_anchor_scaling(self):
        """Test scaling with reference series"""
        # Create sample datetime data
        reference_dates = pd.Series(
            pd.date_range(start="2024-01-01", periods=3, freq="D")
        )
        target_dates = pd.Series(pd.date_range(start="2024-01-02", periods=3, freq="D"))

        # Create scaler with days unit
        scaler = self.scaler_class(unit="D")

        # Set reference series and fit
        scaler.set_reference_time(reference_dates)
        scaler.fit(target_dates)

        # Transform target column
        transformed = scaler.transform(target_dates)

        # Each target date should be 1 day after its reference date
        expected = np.array([1.0, 1.0, 1.0]).reshape(-1, 1)
        np.testing.assert_array_almost_equal(transformed, expected)

        # Test inverse transform
        inverse_transformed = scaler.inverse_transform(transformed)
        expected_dates = target_dates.values.reshape(-1, 1)
        np.testing.assert_array_equal(inverse_transformed, expected_dates)

    def test_time_anchor_seconds(self):
        """Test scaling with seconds unit"""
        # Create sample datetime data with 1 hour differences
        reference_dates = pd.Series(
            pd.date_range(start="2024-01-01", periods=3, freq="h")
        )
        target_dates = pd.Series(
            pd.date_range(start="2024-01-01 01:00", periods=3, freq="h")
        )

        # Create scaler with seconds unit
        scaler = self.scaler_class(unit="S")

        # Set reference series and fit
        scaler.set_reference_time(reference_dates)
        scaler.fit(target_dates)

        # Transform target column
        transformed = scaler.transform(target_dates)

        # Each target date should be 3600 seconds (1 hour) after its reference date
        expected = np.array([3600.0, 3600.0, 3600.0]).reshape(-1, 1)
        np.testing.assert_array_almost_equal(transformed, expected)

        # Test inverse transform
        inverse_transformed = scaler.inverse_transform(transformed)
        expected_dates = target_dates.values.reshape(-1, 1)
        np.testing.assert_array_equal(inverse_transformed, expected_dates)

    def test_missing_reference(self):
        """Test error when reference is not set"""
        dates = pd.Series(pd.date_range(start="2024-01-01", periods=3, freq="D"))

        scaler = self.scaler_class(unit="D")

        with pytest.raises(ValueError, match="Reference series not set"):
            scaler.fit(dates)

    def test_length_mismatch(self):
        """Test error when reference and target lengths don't match"""
        reference_dates = pd.Series(
            pd.date_range(start="2024-01-01", periods=3, freq="D")
        )
        target_dates = pd.Series(pd.date_range(start="2024-01-01", periods=4, freq="D"))

        scaler = self.scaler_class(unit="D")
        scaler.set_reference_time(reference_dates)

        with pytest.raises(
            ValueError, match="Target and reference must have same length"
        ):
            scaler.fit(target_dates)

    def test_invalid_unit(self):
        """Test error with invalid time unit"""
        with pytest.raises(
            ValueError, match="unit must be either 'D'\\(days\\) or 'S'\\(seconds\\)"
        ):
            self.scaler_class(unit="H")

    def test_invalid_data_type(self):
        """Test error with non-datetime data"""
        reference_dates = pd.Series(
            pd.date_range(start="2024-01-01", periods=3, freq="D")
        )
        invalid_data = pd.Series([1, 2, 3])

        scaler = self.scaler_class(unit="D")
        scaler.set_reference_time(reference_dates)

        with pytest.raises(ValueError, match="Data must be in datetime format"):
            scaler.fit(invalid_data)

    def test_invalid_reference_type(self):
        """Test error with non-datetime reference"""
        invalid_reference = pd.Series([1, 2, 3])
        dates = pd.Series(pd.date_range(start="2024-01-01", periods=3, freq="D"))

        scaler = self.scaler_class(unit="D")

        with pytest.raises(ValueError, match="Reference data must be datetime type"):
            scaler.set_reference_time(invalid_reference)



import hashlib
import logging
import os
from abc import ABC, abstractmethod

import requests

from petsard.exceptions import BenchmarkDatasetsError


def digest_sha256(filepath):
    """
    Calculate SHA-256 value of file. Load 128KB at one time.
    ...
    Args:
        filepath (str) Openable file full path.
    ...
    return:
        (str) SHA-256 value of file.
    """
    sha256hash = hashlib.sha256()
    with open(filepath, "rb") as f:
        for byte_block in iter(lambda: f.read(131072), b""):
            sha256hash.update(byte_block)
    return sha256hash.hexdigest()


class BaseBenchmarker(ABC):
    """
    BaseBenchmarker
        Base class for all "Benchmarker".
        The "Benchmarker" class defines the common API
        that all the "Loader" need to implement, as well as common functionality.
    """

    def __init__(self, config: dict):
        """
        Attributes:
            _logger (logging.Logger): The logger object.
            config (dict) The configuration of the benchmarker.
                benchmark_bucket_name (str) The name of the S3 bucket.
                benchmark_filename (str)
                    The name of the benchmark data from benchmark_datasets.yaml.
                benchmark_sha256 (str)
                    The SHA-256 value of the benchmark data from benchmark_datasets.yaml.
                filepath (str) The full path of the benchmark data in local.
                benchmark_already_exist (bool)
                    If the benchmark data already exist. Default is False.
        """
        self._logger: logging.Logger = logging.getLogger(
            f"PETsARD.{self.__class__.__name__}"
        )
        self._logger.info(
            f"Initializing Benchmarker with benchmark_filename: {config['benchmark_filename']}"
        )

        self.config: dict = config
        self.config["benchmark_already_exist"] = False
        if os.path.exists(self.config["filepath"]):
            # if same name data already exist, check the sha256hash,
            #     if match, ignore download and continue,
            #     if NOT match, raise Error
            self._verify_file(already_exist=True)
        else:
            # if same name data didn't exist,
            #     confirm "./benchmark/" folder is exist (create it if not)
            os.makedirs("benchmark", exist_ok=True)

    @abstractmethod
    def download(self):
        """
        Download the data
        """
        raise NotImplementedError()

    def _verify_file(self, already_exist: bool = True):
        """
        Verify the exist file is match to records

        Args:
            already_exist (bool) If the file already exist. Default is True.
              False means verify under download process.
        """
        file_sha256hash = digest_sha256(self.config["filepath"])

        if file_sha256hash == self.config["benchmark_sha256"]:
            self.config["benchmark_already_exist"] = True
        else:
            if already_exist:
                self._logger.error(f"SHA-256 mismatch: {self.config['filepath']}")
                raise BenchmarkDatasetsError(
                    f"SHA-256 mismatch: {self.config['filepath']}. "
                )
            else:
                try:
                    os.remove(self.config["filepath"])
                    self._logger.error(
                        f"Downloaded file SHA-256 mismatch: {self.config['benchmark_filename']} from "
                        f"{self.config['benchmark_bucket_name']}. "
                    )
                    raise BenchmarkDatasetsError(
                        f"Downloaded file SHA-256 mismatch: {self.config['benchmark_filename']} from "
                        f"{self.config['benchmark_bucket_name']}. "
                    )
                except OSError as e:
                    self._logger.error(
                        f"Failed to remove file: {self.config['filepath']}. Please delete it manually."
                    )
                    raise OSError(
                        f"Failed to remove file: {self.config['filepath']}. Please delete it manually."
                    ) from e


class BenchmarkerRequests(BaseBenchmarker):
    """
    BenchmarkerRequests
        Download benchmark dataset via requests.
        Expect for public bucket.

    """

    def __init__(self, config: dict):
        super().__init__(config)

    def download(self) -> None:
        """
        Use requests.get() to download data,
            than confirm its SHA-256 is matched.

        """
        if self.config["benchmark_already_exist"]:
            self._logger.info(f"Using local file: {self.config['filepath']}")
        else:
            url = (
                f"https://"
                f"{self.config['benchmark_bucket_name']}"
                f".s3.amazonaws.com/"
                f"{self.config['benchmark_filename']}"
            )
            self._logger.info(f"Downloading from: {url}")
            with requests.get(url, stream=True) as response:
                if response.status_code == 200:
                    with open(self.config["filepath"], "wb") as f:
                        # load 8KB at one time
                        for chunk in response.iter_content(chunk_size=8192):
                            f.write(chunk)
                    self._logger.info(
                        f"Download completed: {self.config['benchmark_filename']}"
                    )
                else:
                    self._logger.error(
                        f"Download failed: status={response.status_code}, url={url}"
                    )
                    raise BenchmarkDatasetsError(
                        f"Download failed: status={response.status_code}, url={url}"
                    )
            self._verify_file(already_exist=False)



import logging
import re
from copy import deepcopy
from dataclasses import dataclass, field
from enum import Enum, auto
from typing import Any, Optional

import numpy as np
import pandas as pd
from sklearn.cluster import KMeans
from sklearn.ensemble import (
    GradientBoostingClassifier,
    GradientBoostingRegressor,
    RandomForestClassifier,
    RandomForestRegressor,
)
from sklearn.linear_model import LinearRegression, LogisticRegression
from sklearn.metrics import f1_score, silhouette_score
from sklearn.preprocessing import OneHotEncoder, StandardScaler
from sklearn.svm import SVC

from petsard.config_base import BaseConfig
from petsard.evaluator.evaluator_base import BaseEvaluator
from petsard.exceptions import ConfigError, UnsupportedMethodError
from petsard.metadater.types.data_types import safe_round


class MLUtilityMap(Enum):
    """
    Enumeration for MLUtility method mapping.
    """

    REGRESSION: int = auto()
    CLASSIFICATION: int = auto()
    CLUSTER: int = auto()

    @classmethod
    def map(cls, method: str) -> int:
        """
        Get suffixes mapping int value

        Args:
            method (str): evaluating method

        Return:
            (int): The method code.
        """
        return cls.__dict__[re.sub(r"^mlutility-", "", method).upper()]


@dataclass
class MLUtilityConfig(BaseConfig):
    """
    Configuration for the MLUtility Evaluator.

    Attributes:
        eval_method: The method name of how you are evaluating data.
        eval_method_code (Optional[int]): The mapped method code.
        target (str, optional):
            The target column for regression/classification.
            Should be a numerical column for regression.
        n_clusters (list[int], optional):
            List of cluster numbers for clustering. Default is [4, 5, 6].
        REQUIRED_INPUT_KEYS (list[str]): The required keys in the input data.
        n_rows (dict[str, int]): Number of rows in each data.
        category_cols (list[str]): List of categorical columns in the data.
        category_cols_cardinality (dict[str, dict[str, int]]):
            Cardinality of categorical columns in the data.
        _logger (logging.Logger): The logger object.
    """

    eval_method: str
    eval_method_code: Optional[int] = None
    target: Optional[str] = None
    n_clusters: list[int] = field(default_factory=lambda: [4, 5, 6])
    REQUIRED_INPUT_KEYS: list[str] = field(
        default_factory=lambda: ["ori", "syn", "control"]
    )
    n_rows: dict[str, int] = field(default_factory=dict)
    category_cols: list[str] = field(default_factory=list)
    category_cols_cardinality: dict[str, dict[str, int]] = field(default_factory=dict)

    def __post_init__(self):
        super().__post_init__()
        error_msg: Optional[str] = None

        # Map and validate method
        try:
            self.eval_method_code = MLUtilityMap.map(self.eval_method)
        except KeyError:
            error_msg = f"Unsupported method: {self.eval_method}"
            self._logger.error(error_msg)
            raise UnsupportedMethodError(error_msg)

        # Validate n_clusters
        if not isinstance(self.n_clusters, list):
            error_msg = "n_clusters must be a list of integers"
            self._logger.error(error_msg)
            raise ConfigError(error_msg)

        if not all(isinstance(x, int) for x in self.n_clusters):
            error_msg = "All elements in n_clusters must be integers"
            self._logger.error(error_msg)
            raise ConfigError(error_msg)

        # Validate target for specific methods
        if self.eval_method_code in [
            MLUtilityMap.REGRESSION,
            MLUtilityMap.CLASSIFICATION,
        ]:
            if not self.target:
                error_msg = f"Target column is required for {self.eval_method}"
                self._logger.error(error_msg)
                raise ConfigError(error_msg)
        elif self.eval_method_code == MLUtilityMap.CLUSTER and self.target is not None:
            self.target = None
            error_msg = (
                "Target column is not required for clustering, input will be ignored"
            )
            self._logger.info(error_msg)

    def update_data(self, data: dict[str, pd.DataFrame]) -> None:
        """
        Validate input data for MLUtility evaluation.

        Args:
            data (Dict[str, pd.DataFrame]): Input data dictionary.

        Raises:
            ConfigError: If data is invalid.
        """
        error_msg: Optional[str] = None

        # Validate required keys
        if not all(key in data for key in self.REQUIRED_INPUT_KEYS):
            error_msg = f"Missing required keys. Expected: {self.REQUIRED_INPUT_KEYS}"
            self._logger.error(error_msg)
            raise ConfigError(error_msg)

        # Validate data is not empty after removing NaNs
        for key in self.REQUIRED_INPUT_KEYS:
            df = data[key].dropna()
            if df.empty:
                error_msg = f"Data for {key} is empty after removing missing values"
                self._logger.error(error_msg)
                raise ConfigError(error_msg)

        # Validate target column for regression/classification
        if self.eval_method_code in [
            MLUtilityMap.REGRESSION,
            MLUtilityMap.CLASSIFICATION,
        ]:
            for key in self.REQUIRED_INPUT_KEYS:
                if self.target not in data[key].columns:
                    error_msg = f"Target column '{self.target}' not found in {key} data"
                    self._logger.error(error_msg)
                    raise ConfigError(error_msg)

        self.n_rows = {key: data[key].shape[0] for key in self.REQUIRED_INPUT_KEYS}
        self._logger.debug(f"Number of rows in each data: {self.n_rows}")

        self.category_cols: list[str] = [
            col
            for col in data["ori"].columns
            if not pd.api.types.is_numeric_dtype(data["ori"][col])
        ]
        self._logger.debug(f"Category columns: {self.category_cols}")
        self.category_cols_cardinality = {
            key: {col: data[key][col].nunique() for col in self.category_cols}
            for key in self.REQUIRED_INPUT_KEYS
        }
        self._logger.debug(
            f"Category columns cardinality: {self.category_cols_cardinality}"
        )


class MLUtility(BaseEvaluator):
    """
    Evaluator for MLUtility method.
    """

    REQUIRED_INPUT_KEYS: list[str] = ["ori", "syn", "control"]
    LOWER_BOUND_MAP: dict[int, float] = {
        MLUtilityMap.CLASSIFICATION: 0.0,
        MLUtilityMap.CLUSTER: -1.0,
        MLUtilityMap.REGRESSION: 0.0,
    }
    AVAILABLE_SCORES_GRANULARITY: list[str] = ["global", "details"]
    HIGH_CARDINALITY_THRESHOLD: float = 0.1  # 1/10 of rows
    RANDOM_STATE_SEED: int = 42

    def __init__(self, config: dict):
        """
        Args:
            config (dict): A dictionary containing the configuration settings.

        Attr:
            REQUIRED_INPUT_KEYS (list[str]): The required keys in the input data.
            LOWER_BOUND_MAP (dict[int, float]): The lower bound for the score.
            AVAILABLE_SCORES_GRANULARITY (list[str]): The available scores granularity.
            HIGH_CARDINALITY_THRESHOLD (float):
                The threshold for high cardinality. Represents as percentage of rows.
            RANDOM_STATE_SEED (int): The random state seed for the models.
            _logger (logging.Logger): The logger object.
            config (dict): A dictionary containing the configuration settings.
            mlutility_config (MLUtilityConfig): The configuration object.
            _impl (Optional[dict[str, callable]]): The evaluator object.
                - 'ori': The evaluator object for the original data.
                - 'syn': The evaluator object for the synthetic data.
        """
        super().__init__(config=config)
        self._logger: logging.Logger = logging.getLogger(
            f"PETsARD.{self.__class__.__name__}"
        )

        self.MLUTILITY_CLASS_MAP: dict[int, callable] = {
            MLUtilityMap.CLASSIFICATION: self._classification,
            MLUtilityMap.CLUSTER: self._cluster,
            MLUtilityMap.REGRESSION: self._regression,
        }

        self._logger.debug(f"Verifying MLUtilityConfig with parameters: {self.config}")
        self.mlutility_config = MLUtilityConfig(**self.config)
        self._logger.debug("MLUtilityConfig successfully initialized")

        self._impl: Optional[dict[str, callable]] = None

    def _preprocessing(
        self, data: dict[str, pd.DataFrame]
    ) -> tuple[dict[str, pd.DataFrame], str]:
        """
        Preprocess the data for the evaluation.
            1. Remove missing values.
            2. Remove high cardinality columns.
            3. One-hot encoding for categorical columns.
            4. Standardize the data.
            5. Check if the target column is constant.

        Args:
            data (dict[str, pd.DataFrame]): The data to be evaluated.

        Returns:
            (tuple[dict[str, pd.DataFrame], str]):
                The preprocessed data and the status of the preprocessing.
                - (dict[str, pd.DataFrame]): The preprocessed data.
                    - (str): key of the data.
                        - X (pd.DataFrame): The feature columns.
                        - y (pd.Series): The target column.
                            Only for regression and classification.
                - (str): The status of the preprocessing.
                    - 'success': The preprocessing is successful.
                    - 'missing_values_cause_empty':
                        The data is empty after removing missing values.
                    - 'target_is_constant':
                        The target column is constant
        """
        error_msg: Optional[str] = None
        inprogress_data: dict[str, pd.DataFrame] = deepcopy(data)
        data_now: Optional[pd.DataFrame] = None
        preprocessed_data: dict[str, pd.DataFrame] = {}

        for key in self.REQUIRED_INPUT_KEYS:
            data_now = inprogress_data[key].copy()
            preprocessed_data[key] = {}

            data_now.dropna(how="any")
            self._logger.debug(f"Missing values removed for {key}")
            if data_now.shape[0] == 0:
                error_msg = f"The data is empty after removing missing values for {key}"
                self._logger.warnings(error_msg)
                return None, "missing_values_cause_empty"

        temp_category_cols: list[str] = (
            [
                col
                for col in self.mlutility_config.category_cols
                if col != self.mlutility_config.target
            ]
            if self.mlutility_config.eval_method_code
            in [
                MLUtilityMap.CLASSIFICATION,
                MLUtilityMap.REGRESSION,
            ]
            else self.mlutility_config.category_cols
        )
        self._logger.debug(
            f"Category columns included: {self.mlutility_config.category_cols}"
        )
        if len(temp_category_cols) != 0:
            for col in self.mlutility_config.category_cols:
                # if the cardinality of the column is too high, remove the column
                if (
                    self.mlutility_config.category_cols_cardinality["ori"][col]
                    >= round(
                        self.mlutility_config.n_rows["ori"]
                        * self.HIGH_CARDINALITY_THRESHOLD
                    )
                ) or (
                    self.mlutility_config.category_cols_cardinality["syn"][col]
                    >= round(
                        self.mlutility_config.n_rows["syn"]
                        * self.HIGH_CARDINALITY_THRESHOLD
                    )
                ):
                    error_msg = (
                        f"The cardinality of the column {col} is too high. "
                        f"Ori: Over row numbers {self.mlutility_config.n_rows['ori']},"
                        f" column cardinality {self.mlutility_config.category_cols_cardinality['ori'][col]}. "
                        f"Syn: Over row numbers {self.mlutility_config.n_rows['syn']},"
                        f" column cardinality {self.mlutility_config.category_cols_cardinality['syn'][col]}. "
                        f"The column {col} is removed."
                    )
                    self._logger.warning(error_msg)
                    temp_category_cols.remove(col)
                    for key in self.REQUIRED_INPUT_KEYS:
                        inprogress_data[key] = inprogress_data[key].drop(
                            columns=col, inplace=False
                        )
                        self._logger.debug(f"Column {col} removed for {key}")

            # One-hot encoding
            ohe = OneHotEncoder(
                drop="first",
                sparse_output=False,
                handle_unknown="infrequent_if_exist",
            )
            ohe.fit(
                pd.concat(
                    [
                        inprogress_data["ori"][temp_category_cols],
                        inprogress_data["syn"][temp_category_cols],
                        inprogress_data["control"][temp_category_cols],
                    ]
                )
            )
            self._logger.debug("One-hot encoding completed")
            for key in self.REQUIRED_INPUT_KEYS:
                data_now = inprogress_data[key].copy()
                data_now = pd.DataFrame(
                    ohe.transform(data_now[temp_category_cols]),
                    columns=ohe.get_feature_names_out(temp_category_cols),
                    index=data_now.index,
                )
                inprogress_data[key] = inprogress_data[key].drop(
                    columns=temp_category_cols,
                    inplace=False,
                )
                inprogress_data[key] = pd.concat(
                    [inprogress_data[key], data_now], axis=1
                )
                self._logger.debug(f"One-hot encoding completed for {key}")

        if self.mlutility_config.eval_method_code in [
            MLUtilityMap.CLASSIFICATION,
            MLUtilityMap.REGRESSION,
        ]:
            target: str = self.mlutility_config.target

            if self.mlutility_config.eval_method_code == MLUtilityMap.REGRESSION:
                ss_y = StandardScaler()
                ss_y.fit(
                    np.concatenate(
                        [
                            inprogress_data["ori"][target].values.reshape(-1, 1),
                            inprogress_data["syn"][target].values.reshape(-1, 1),
                            inprogress_data["control"][target].values.reshape(-1, 1),
                        ]
                    )
                )
            ss_X = StandardScaler()
            ss_X.fit(
                pd.concat(
                    [
                        inprogress_data["ori"].drop(columns=[target], inplace=False),
                        inprogress_data["syn"].drop(columns=[target], inplace=False),
                        inprogress_data["control"].drop(
                            columns=[target], inplace=False
                        ),
                    ]
                )
            )

            for key in self.REQUIRED_INPUT_KEYS:
                target_value: np.ndarray = inprogress_data[key][target].values

                # check if the target is constant
                if len(np.unique(target_value)) == 1:
                    error_msg = f"The target column '{target}' is constant."
                    self._logger.warning(error_msg)
                    return None, "target_is_constant"

                preprocessed_data[key]["y"] = (
                    ss_y.transform(target_value.reshape(-1, 1)).ravel()
                    if self.mlutility_config.eval_method_code == MLUtilityMap.REGRESSION
                    else target_value
                )
                preprocessed_data[key]["X"] = ss_X.transform(
                    inprogress_data[key].drop(columns=[target], inplace=False)
                )

        elif self.mlutility_config.eval_method_code == MLUtilityMap.CLUSTER:
            united_columns = list(
                set(
                    inprogress_data["ori"].columns.tolist()
                    + inprogress_data["syn"].columns.tolist()
                    + inprogress_data["control"].columns.tolist()
                )
            )

            ss = StandardScaler()
            ss.fit(
                pd.concat(
                    [
                        inprogress_data["ori"].reindex(
                            columns=united_columns, fill_value=0.0
                        ),
                        inprogress_data["syn"].reindex(
                            columns=united_columns, fill_value=0.0
                        ),
                        inprogress_data["control"].reindex(
                            columns=united_columns, fill_value=0.0
                        ),
                    ],
                    axis=0,
                )
            )

            for key in self.REQUIRED_INPUT_KEYS:
                preprocessed_data[key] = {}
                preprocessed_data[key]["X"] = ss.transform(
                    inprogress_data[key].reindex(
                        columns=united_columns, fill_value=0.0
                    ),
                )

        self._logger.debug(
            f"Encoding and standardization of {self.mlutility_config.eval_method} completed"
        )

        self._logger.debug(
            f"Data preprocessing of {self.mlutility_config.eval_method} completed"
        )

        return preprocessed_data, "success"

    def _adjust_to_lower_bound(self, value: float) -> float:
        """
        Check if the score is beyond the lower bound.
            If the score is less than the lower bound, return the lower bound.
             Otherwise, return the value.
            - For regression and classification, the lower bound is 0.
            - For clustering, the lower bound is -1.

        Args:
            value (float): The value to be checked.
            type (str): The type of the evaluation.

        Returns:
            (float): The value in the range.
        """
        error_msg: Optional[str] = None
        lower_bound: float = self.LOWER_BOUND_MAP[
            self.mlutility_config.eval_method_code
        ]

        if value < lower_bound:
            error_msg = (
                f"The score {value} is less than the lower bound, "
                "indicating the performance is arbitrarily poor. "
                "So the score is set to the lower bound."
            )
            self._logger.warning(error_msg)
            value = lower_bound

        return value

    def _classification(
        self,
        X_train: np.ndarray,
        y_train: np.ndarray,
        X_test: np.ndarray,
        y_test: np.ndarray,
    ) -> dict[str, float]:
        """
        Classification model fitting, evaluation, and testing.
            The models used are logistic regression, SVC, random forest, and gradient boosting.

        The metric used for evaluation is f1 score.

        Args:
            X_train (np.ndarray): The data to be fitted.
            y_train (np.ndarray): The target column of the training data.
            X_test (np.ndarray): The data to be tested.
            y_test (np.ndarray): The target column of the testing data.

        Returns:
            (dict[str, float]): The result of the evaluation.
                - 'logistic_regression': The f1 score of the logistic regression model.
                - 'svc': The f1 score of the SVC model.
                - 'random_forest': The f1 score of the random forest model.
                - 'gradient_boosting': The f1 score of the gradient boosting model
        """
        result: dict[str, float] = {}
        average_method: str = "micro"

        models_to_evaluate_map: dict[str, callable] = {
            "logistic_regression": LogisticRegression(
                random_state=self.RANDOM_STATE_SEED
            ),
            "svc": SVC(random_state=self.RANDOM_STATE_SEED),
            "random_forest": RandomForestClassifier(
                random_state=self.RANDOM_STATE_SEED
            ),
            "gradient_boosting": GradientBoostingClassifier(
                random_state=self.RANDOM_STATE_SEED
            ),
        }

        # train
        for model in models_to_evaluate_map.values():
            model.fit(X_train, y_train)

        # evaluate
        result = {
            name: f1_score(
                y_test,
                model.fit(X_train, y_train).predict(X_test),
                average=average_method,
            )
            for name, model in models_to_evaluate_map.items()
        }

        return result

    def _cluster(self, X_train: np.ndarray, X_test: np.ndarray) -> dict[str, float]:
        """
        Clustering model fitting, evaluation, and testing.
            The models used are KMeans with different number of clusters.
            For the robustness of the results,
                the model is trained and evaluated 5 times,
                and the average of the results is used as the final result.

        The metric used for evaluation is silhouette score.

        Args:
            X_train (pd.DataFrame): The data to be fitted.
            X_test (pd.DataFrame): The data to be tested.
            n_clusters (list): A list of numbers of clusters for clustering.

        Returns:
            result (dict): The result of the evaluation.
        """
        silhouette_score_value: float = None
        result: dict[str, float] = {}

        for k in self.mlutility_config.n_clusters:
            k_model = KMeans(
                random_state=self.RANDOM_STATE_SEED, n_clusters=k, n_init="auto"
            )

            k_model.fit(X_train)

            try:
                silhouette_score_value = silhouette_score(
                    X_test, k_model.predict(X_test)
                )
            except ValueError as e:
                error_msg: str = (
                    "There is only one cluster in the prediction, "
                    "or the valid data samples are too few, "
                    "indicating the performance is arbitrarily poor. "
                    "The score is set to the lower bound."
                    f"Error message: {str(e)}"
                )
                self._logger.warning(error_msg)
                silhouette_score_value = -1

            result[f"KMeans_cluster{k}"] = self._adjust_to_lower_bound(
                silhouette_score_value,
            )

        return result

    def _regression(
        self,
        X_train: np.ndarray,
        y_train: np.ndarray,
        X_test: np.ndarray,
        y_test: np.ndarray,
    ) -> dict[str, float]:
        """
        Regression model fitting, evaluation, and testing.
            The models used are linear regression, random forest, and gradient boosting.

        The metric used for evaluation is R^2.

        Args:
            X_train (np.ndarray): The data to be fitted.
            y_train (np.ndarray): The target column of the training data.
            X_test (np.ndarray): The data to be tested.
            y_test (np.ndarray): The target column of the testing data.

        Returns:
            (dict[str, float]): The result of the evaluation.
                - 'linear_regression': The R^2 score of the linear regression model.
                - 'random_forest': The R^2 score of the random forest model.
                - 'gradient_boosting': The R^2 score of the gradient boosting model
        """
        result: dict[str, float] = {}

        models_to_evaluate_map: dict[str, callable] = {
            "linear_regression": LinearRegression(),
            "random_forest": RandomForestRegressor(random_state=self.RANDOM_STATE_SEED),
            "gradient_boosting": GradientBoostingRegressor(
                random_state=self.RANDOM_STATE_SEED
            ),
        }

        result = {
            name: self._adjust_to_lower_bound(
                model.fit(X_train, y_train).score(X_test, y_test)
            )
            for name, model in models_to_evaluate_map.items()
        }

        return result

    def _get_global(
        self, result: dict[str, Optional[dict[str, float]]]
    ) -> pd.DataFrame:
        """
        Get the global result of the evaluation.

        Args:
            result (dict[str, Optional[dict[str, float]]]): The evaluation result
                - 'ori': The evaluation result of the original data.
                - 'syn': The evaluation result of the synthetic data.

        Returns:
            (pd.DataFrame): The global result of the evaluation.
        """
        ori_value: list[float] = list(result.get("ori", {}).values() or [0])
        syn_value: list[float] = list(result.get("syn", {}).values() or [0])

        compare_df: pd.DataFrame = pd.DataFrame(
            {
                "ori_mean": safe_round(np.mean(ori_value)),
                "ori_std": safe_round(np.std(ori_value)),
                "syn_mean": safe_round(np.mean(syn_value)),
                "syn_std": safe_round(np.std(syn_value)),
            },
            index=[0],
        )

        compare_df["diff"] = safe_round(compare_df["syn_mean"] - compare_df["ori_mean"])

        return compare_df

    def _eval(self, data: dict[str, pd.DataFrame]) -> dict:
        """
        Evaluate the data with the method given in the config.

        Args:
            data (dict[str, pd.DataFrame]): The data to be evaluated.

        Returns:
            (dict) he evaluation result.
        """
        preprocessed_data: dict[str, pd.DataFrame] = {}
        result_details: dict[str, Optional[dict[str, float]]] = {}

        self.mlutility_config.update_data(data)

        preprocessed_data, status = self._preprocessing(data)
        self._logger.debug("Data preprocessing completed")
        if status in ("missing_values_cause_empty", "target_is_constant"):
            self._logger.warning(
                f"Evaluator Preprocessing status: {status}, result will be NaN"
            )
            result_details = {
                "ori": {"error": np.nan},
                "syn": {"error": np.nan},
            }
        else:
            self._logger.debug(
                f"Initializing evaluator with method: {self.config['eval_method']}"
            )
            evaluator_class: Any = self.MLUTILITY_CLASS_MAP[
                self.mlutility_config.eval_method_code
            ]
            self._logger.debug(
                f"Mapped method code: {self.mlutility_config.eval_method_code}"
            )
            self._impl = {"ori": evaluator_class, "syn": evaluator_class}

            result_details = {
                data_type: evaluator_class(
                    **(
                        dict(
                            X_train=preprocessed_data[data_type]["X"],
                            y_train=preprocessed_data[data_type]["y"],
                            X_test=preprocessed_data["control"]["X"],
                            y_test=preprocessed_data["control"]["y"],
                        )
                        if self.mlutility_config.eval_method_code
                        in [MLUtilityMap.CLASSIFICATION, MLUtilityMap.REGRESSION]
                        else dict(
                            X_train=preprocessed_data[data_type]["X"],
                            X_test=preprocessed_data["control"]["X"],
                        )
                    )
                )
                for data_type in ["ori", "syn"]
            }

        return {
            "global": self._get_global(result_details),
            "details": result_details,
        }



import itertools

import numpy as np
import pandas as pd
import pytest

from petsard.constrainer.field_combination_constrainer import (
    FieldCombinationConstrainer,
)
from petsard.exceptions import ConfigError


class TestFieldCombinationConstrainer:
    @pytest.fixture
    def sample_df(self):
        """
        Generate sample data for testing

        Returns:
            pd.DataFrame: DataFrame containing all possible combinations of:
                - Job categories (including NA)
                - Experience levels (including NA)
                - Salary ranges (including NA)
                - Performance grades (including NA)
        """
        jobs = ["Engineer", "Analyst", "Researcher", np.nan]
        levels = ["Junior", "Senior", "Staff", np.nan]
        salaries = [50000, 60000, 70000, np.nan]
        grades = ["A", "B", "C", np.nan]

        data = [
            {
                "job": job,
                "level": level,
                "salary": salary,
                "grade": grade,
            }
            for job, level, salary, grade in itertools.product(
                jobs, levels, salaries, grades
            )
        ]

        return pd.DataFrame(data)

    def test_validate_config_existing_columns(self, sample_df):
        """Test validate_config method with existing columns"""
        constraints = [
            ({"level": "salary"}, {"Senior": 70000}),
            ({("job", "level"): "grade"}, {("Engineer", "Senior"): "A"}),
        ]

        constrainer = FieldCombinationConstrainer(constraints)
        try:
            constrainer.validate_config(sample_df)
        except ConfigError:
            pytest.fail("Valid columns raised unexpected ConfigError")

    def test_validate_config_nonexistent_columns(self, sample_df):
        """Test validate_config method with non-existent columns"""
        constraints = [
            ({"nonexistent_field": "salary"}, {"Senior": 70000}),
            ({("job", "nonexistent_level"): "grade"}, {("Engineer", "Senior"): "A"}),
        ]

        constrainer = FieldCombinationConstrainer(constraints)

        with pytest.raises(
            ConfigError, match="Columns .* do not exist in the DataFrame"
        ):
            constrainer.validate_config(sample_df)

    def test_apply_with_nonexistent_columns(self, sample_df):
        """Test apply method with non-existent columns"""
        constraints = [({"nonexistent_field": "salary"}, {"Senior": 70000})]

        constrainer = FieldCombinationConstrainer(constraints)

        with pytest.raises(
            ConfigError, match="Columns .* do not exist in the DataFrame"
        ):
            constrainer.apply(sample_df)

    def test_single_field_constraint_with_specific_value(self, sample_df):
        """Test single field constraint with specific value"""
        constraints = [({"level": "salary"}, {"Senior": 75000})]

        constrainer = FieldCombinationConstrainer(constraints)
        result = constrainer.apply(sample_df)

        # Verify only Senior level with 70000 salary is kept
        assert len(result) == 192  # 4*4*4*4 - 4*4*4
        assert (result[result["level"] == "Senior"]["salary"] == 70000).all()

    def test_valid_single_field_constraint(self):
        """Test valid single field constraint configuration"""
        constraints = [
            ({"level": "salary"}, {"Senior": 70000}),
            ({"job": "grade"}, {"Engineer": "A"}),
        ]

        try:
            FieldCombinationConstrainer(constraints)
        except ConfigError:
            pytest.fail("Valid single field constraint raised unexpected ConfigError")

    def test_valid_multi_field_constraint(self):
        """Test valid multi-field constraint configuration"""
        constraints = [
            ({("job", "level"): "salary"}, {("Engineer", "Junior"): 70000}),
            ({("job", "level"): "grade"}, {("Engineer", "Senior"): "A"}),
            (
                {("grade", "job", "level"): "salary"},
                {("A", "Engineer", "Senior"): [70000]},
            ),
        ]

        try:
            FieldCombinationConstrainer(constraints)
        except ConfigError:
            pytest.fail("Valid multi-field constraint raised unexpected ConfigError")

    def test_invalid_constraints_not_list(self):
        """Test that non-list constraints raise ConfigError"""
        with pytest.raises(ConfigError, match="Constraints must be a list"):
            FieldCombinationConstrainer({"department": "salary"})

    def test_invalid_constraint_structure(self):
        """Test invalid constraint tuple structure"""
        with pytest.raises(
            ConfigError, match="Each constraint must be a tuple with two elements"
        ):
            FieldCombinationConstrainer([("invalid",)])
        with pytest.raises(ConfigError):
            FieldCombinationConstrainer(
                [
                    {("grade", "job", "level"): "salary"},
                    {("A", "Engineer", "Senior"): [70000]},
                ]
            )

    def test_invalid_field_map(self):
        """Test invalid field map"""
        with pytest.raises(
            ConfigError,
            match="Field map must be a dictionary with exactly one key-value pair",
        ):
            FieldCombinationConstrainer([({}, {"Senior": 70000})])
        with pytest.raises(
            ConfigError,
            match="Field map must be a dictionary with exactly one key-value pair",
        ):
            FieldCombinationConstrainer([({1: 2, 3: 4}, {"Senior": 70000})])

    def test_invalid_source_fields(self):
        """Test invalid source fields type"""
        with pytest.raises(
            ConfigError, match="Source fields must be a string or tuple of strings"
        ):
            FieldCombinationConstrainer([({1: "salary"}, {"Senior": 70000})])
        with pytest.raises(
            ConfigError, match="Source fields must be a string or tuple of strings"
        ):
            FieldCombinationConstrainer(
                [({("job", 1): "salary"}, {("Engineer", "Junior"): 70000})]
            )

    def test_invalid_target_field(self):
        """Test invalid target field type"""
        with pytest.raises(ConfigError, match="Target field must be a string"):
            FieldCombinationConstrainer(
                [({("job", "level"): 1}, {("Engineer", "Junior"): 70000})]
            )

    def test_multi_field_source_value_length_mismatch(self):
        """Test mismatch between multi-field source fields and source values"""
        with pytest.raises(ConfigError, match="Source value must be a tuple of length"):
            FieldCombinationConstrainer(
                [({("job", "level"): "salary"}, {("Engineer",): 75000})]
            )



import numpy as np
import pandas as pd
import pytest

from petsard.exceptions import UnfittedError
from petsard.processor.encoder import EncoderLabel, EncoderUniform


class Test_EncoderUniform:
    def test_EncoderUniform(self):
        # Prepare test data
        df_data1 = ["A"] * 7 + ["B"] * 3 + ["C"] * 5 + ["D"] * 5
        df_data2 = [0] * 20
        df_data = pd.DataFrame({"col1": pd.Categorical(df_data1), "col2": df_data2})

        # Data for triggering error
        df_data1_error = ["A"] * 7 + ["B"] * 3 + ["E"] * 5 + ["D"] * 5
        df_data_error = pd.DataFrame({"col1": df_data1_error, "col2": df_data2})

        # Create an instance of the class
        encoder = EncoderUniform()

        with pytest.raises(UnfittedError):
            encoder.transform(df_data)

        with pytest.raises(UnfittedError):
            encoder.inverse_transform(df_data)

        # Call the method to be tested
        encoder.fit(df_data["col1"])

        assert set(encoder.labels) == set(["A", "B", "C", "D"])
        assert len(encoder.lower_values) == 4
        assert len(encoder.upper_values) == 4
        assert encoder.lower_values[2] == encoder.upper_values[1]

        with pytest.raises(ValueError):
            encoder.transform(df_data_error["col1"])

        transformed = encoder.transform(df_data["col1"])

        print(transformed)

        assert isinstance(transformed, np.ndarray)
        assert transformed.dtype == np.float64
        assert transformed.min() >= 0
        assert transformed.max() <= 1

        # Data for triggering error
        transformed_error = transformed.copy()
        transformed_error[0] = 2.9

        with pytest.raises(ValueError):
            encoder.inverse_transform(transformed_error)

        rtransformed = encoder.inverse_transform(transformed)

        assert list(rtransformed) == list(df_data["col1"].values)


class Test_EncoderLabel:
    def test_EncoderLabel(self):
        # Prepare test data
        df_data1 = ["A"] * 7 + ["B"] * 3 + ["C"] * 5 + ["D"] * 5
        df_data2 = [0] * 20
        df_data = pd.DataFrame({"col1": pd.Categorical(df_data1), "col2": df_data2})

        # Data for triggering error
        df_data1_error = ["A"] * 7 + ["B"] * 3 + ["E"] * 5 + ["D"] * 5
        df_data_error = pd.DataFrame({"col1": df_data1_error, "col2": df_data2})

        # Create an instance of the class
        encoder = EncoderLabel()

        with pytest.raises(UnfittedError):
            encoder.transform(df_data)

        with pytest.raises(UnfittedError):
            encoder.inverse_transform(df_data)

        # Call the method to be tested
        encoder.fit(df_data["col1"])

        assert set(encoder.labels) == set(["A", "B", "C", "D"])

        with pytest.raises(ValueError):
            encoder.transform(df_data_error["col1"])

        transformed = encoder.transform(df_data["col1"])

        print(transformed)

        assert isinstance(transformed, np.ndarray)
        assert transformed.dtype == int

        rtransformed = encoder.inverse_transform(transformed)

        assert list(rtransformed) == list(df_data["col1"].values)



import numpy as np
import pandas as pd

from petsard.metadater import Metadata


class MySynthesizer_Shuffle:
    """
    A simple synthesizer that shuffles each column independently.

    This synthesizer preserves the distribution of each column while breaking
    the relationships between columns. It can be useful for simple anonymization
    or as a baseline synthetic data generation method.
    """

    def __init__(self, config: dict, metadata: Metadata):
        """
        Initialize the synthesizer.

        In this demo we don't use metadata, but please keep it in the signature,
        and feel free to prepare your synthesizer to use it.

        Args:
            metadata (Metadata): The metadata object.
        """
        self.config: dict = config
        self.result: pd.DataFrame = None

        if "random_seed" in self.config:
            np.random.seed(self.config["random_seed"])

    def fit(self, data: pd.DataFrame) -> None:
        """
        Shuffle Algorithm
        """
        if not isinstance(data, pd.DataFrame):
            raise TypeError("Data must be a pandas DataFrame")

        if data.empty:
            raise ValueError("Cannot fit on empty data")

        columns: list[str] = data.columns.tolist()
        synthetic_data: pd.DataFrame = pd.DataFrame()

        values: np.ndarray = None
        for col in columns:
            # Get the column data and handle different types appropriately
            original_series: pd.Series = data[col]

            if pd.api.types.is_categorical_dtype(original_series):
                # For categorical data, convert to codes, shuffle, then map back
                codes = original_series.cat.codes.values.copy()
                np.random.shuffle(codes)
                # Convert shuffled codes back to categorical values
                synthetic_data[col] = pd.Categorical.from_codes(
                    codes,
                    categories=original_series.cat.categories,
                    ordered=original_series.cat.ordered,
                )
            else:
                # For non-categorical data, create a copy of values and shuffle
                values = original_series.values.copy()
                np.random.shuffle(values)
                synthetic_data[col] = values

        self.result = synthetic_data

    def sample(self) -> pd.DataFrame:
        return self.result



"""Field-level type definitions"""

from dataclasses import dataclass, field
from datetime import datetime
from typing import Any, Dict, List, Optional, Tuple, Union

from petsard.metadater.types.data_types import DataType, LogicalType


@dataclass(frozen=True)
class FieldStats:
    """
    Immutable statistics for a field

    Attributes:
        row_count: Total number of rows
        na_count: Number of null/missing values
        na_percentage: Percentage of null values
        distinct_count: Number of unique values
        min_value: Minimum value (for numeric types)
        max_value: Maximum value (for numeric types)
        mean_value: Mean value (for numeric types)
        std_value: Standard deviation (for numeric types)
        quantiles: Quantile values (for numeric types)
        most_frequent: Most frequent values and their counts
    """

    row_count: int = 0
    na_count: int = 0
    na_percentage: float = 0.0
    distinct_count: int = 0
    min_value: Optional[Union[int, float]] = None
    max_value: Optional[Union[int, float]] = None
    mean_value: Optional[float] = None
    std_value: Optional[float] = None
    quantiles: Optional[Dict[float, Union[int, float]]] = None
    most_frequent: Optional[List[Tuple[Any, int]]] = None


@dataclass(frozen=True)
class FieldConfig:
    """
    Immutable configuration for field processing

    Attributes:
        type_hint: Hint for data type inference
        logical_type: Logical type for the field
        nullable: Whether the field can contain null values
        description: Human-readable description
        cast_error: Error handling strategy ('raise', 'coerce', 'ignore')
        auto_detect_leading_zeros: Whether to automatically detect and preserve leading zeros
        force_nullable_integers: Whether to force use of nullable integer types
        properties: Additional field properties
    """

    type_hint: Optional[str] = None
    logical_type: Optional[Union[LogicalType, str]] = None
    nullable: Optional[bool] = None
    description: Optional[str] = None
    cast_error: str = "coerce"
    auto_detect_leading_zeros: bool = True
    force_nullable_integers: bool = True
    properties: Dict[str, Any] = field(default_factory=dict)

    def __post_init__(self):
        """Validate configuration after initialization"""
        if self.cast_error not in ["raise", "coerce", "ignore"]:
            raise ValueError(f"Invalid cast_error: {self.cast_error}")


@dataclass(frozen=True)
class FieldMetadata:
    """
    Immutable metadata for a single field

    Attributes:
        name: Field name
        data_type: Basic data type
        logical_type: Logical/semantic type
        nullable: Whether field can contain nulls
        source_dtype: Original pandas dtype
        target_dtype: Optimized target dtype
        description: Field description
        cast_error: Error handling strategy
        stats: Field statistics
        properties: Additional properties
        created_at: Creation timestamp
        updated_at: Last update timestamp
    """

    name: str
    data_type: DataType
    logical_type: Optional[LogicalType] = None
    nullable: bool = True
    source_dtype: Optional[str] = None
    target_dtype: Optional[str] = None
    description: Optional[str] = None
    cast_error: str = "coerce"
    stats: Optional[FieldStats] = None
    properties: Dict[str, Any] = field(default_factory=dict)
    created_at: datetime = field(default_factory=datetime.now)
    updated_at: datetime = field(default_factory=datetime.now)

    def with_stats(self, stats: FieldStats) -> "FieldMetadata":
        """Create a new FieldMetadata with updated stats"""
        return FieldMetadata(
            name=self.name,
            data_type=self.data_type,
            logical_type=self.logical_type,
            nullable=self.nullable,
            source_dtype=self.source_dtype,
            target_dtype=self.target_dtype,
            description=self.description,
            cast_error=self.cast_error,
            stats=stats,
            properties=self.properties,
            created_at=self.created_at,
            updated_at=datetime.now(),
        )

    def with_target_dtype(self, target_dtype: str) -> "FieldMetadata":
        """Create a new FieldMetadata with updated target dtype"""
        return FieldMetadata(
            name=self.name,
            data_type=self.data_type,
            logical_type=self.logical_type,
            nullable=self.nullable,
            source_dtype=self.source_dtype,
            target_dtype=target_dtype,
            description=self.description,
            cast_error=self.cast_error,
            stats=self.stats,
            properties=self.properties,
            created_at=self.created_at,
            updated_at=datetime.now(),
        )

    def with_logical_type(self, logical_type: LogicalType) -> "FieldMetadata":
        """Create a new FieldMetadata with updated logical type"""
        return FieldMetadata(
            name=self.name,
            data_type=self.data_type,
            logical_type=logical_type,
            nullable=self.nullable,
            source_dtype=self.source_dtype,
            target_dtype=self.target_dtype,
            description=self.description,
            cast_error=self.cast_error,
            stats=self.stats,
            properties=self.properties,
            created_at=self.created_at,
            updated_at=datetime.now(),
        )


# Type aliases
FieldConfigDict = Dict[str, Any]
FieldMetadataDict = Dict[str, FieldMetadata]
